"""
Test script for the hierarchical classes.
"""

import os

import numpy as np
import pytest
from bilby.core.grid import Grid
from bilby.core.prior import DirichletPriorDict, Uniform
from bilby.core.result import Result, ResultList
from cwinpy.hierarchical import (
    BaseDistribution,
    BoundedGaussianDistribution,
    DeltaFunctionDistribution,
    ExponentialDistribution,
    MassQuadrupoleDistribution,
    PowerLawDistribution,
    create_distribution,
)


class TestDistributionObjects(object):
    """
    Tests for the distribution objects.
    """

    def test_base_distribution(self):
        """
        Test the BaseDistribution object.
        """

        name = "test"

        # test failure for unknown distribution
        with pytest.raises(ValueError):
            BaseDistribution(name, "kjsgdkdgkjgsda")

        # test failure for inappropriate bounds
        with pytest.raises(ValueError):
            BaseDistribution(name, "gaussian", low=0.0, high=-1.0)

        # test failure for unknown hyperparameter name
        with pytest.raises(KeyError):
            hyper = {"mu": [1], "dkgwkufd": [2]}
            BaseDistribution(name, "gaussian", hyperparameters=hyper)

        # test failure with invalid hyperparameter type
        with pytest.raises(TypeError):
            BaseDistribution(name, "gaussian", hyperparameters="blah")

        with pytest.raises(TypeError):
            BaseDistribution(name, "exponential", hyperparameters="blah")

        # test default log_pdf is NaN
        hyper = {"mu": 2.0}
        dist = BaseDistribution(name, "exponential", hyperparameters=hyper)

        assert dist["mu"] == hyper["mu"]
        assert np.isnan(dist.log_pdf({}, 0))
        assert dist.sample({}) is None

        # test failure when getting unknown item
        with pytest.raises(KeyError):
            _ = dist["kgksda"]

        del dist

        # test setter failure
        dist = BaseDistribution(name, "exponential")

        with pytest.raises(KeyError):
            dist["madbks"] = Uniform(0.0, 1.0, "mu")

        # test setter
        dist["mu"] = Uniform(0.0, 1.0, "mu")
        assert isinstance(dist["mu"], Uniform)

        # test failure with invalid hyperparameter type
        with pytest.raises(TypeError):
            hyper = "blah"
            BaseDistribution(name, "deltafunction", hyperparameters=hyper)

    def test_bounded_gaussian(self):
        """
        Test the BoundedGaussianDistribution class.
        """

        name = "test"

        # test failures
        with pytest.raises(TypeError):
            BoundedGaussianDistribution(name, mus="blah")

        with pytest.raises(TypeError):
            BoundedGaussianDistribution(name, mus=[1], sigmas="blah")

        with pytest.raises(TypeError):
            BoundedGaussianDistribution(name, mus=[1], sigmas=[1], weights="blah")

        with pytest.raises(TypeError):
            BoundedGaussianDistribution(name, mus=[1], sigmas=[1], weights=23.4)

        with pytest.raises(ValueError):
            BoundedGaussianDistribution(name, mus=[1.0], sigmas=[1.0, 2.0])

        with pytest.raises(ValueError):
            BoundedGaussianDistribution(
                name, mus=[1.0, 2.0], sigmas=[1.0, 2.0], weights=[1]
            )

        with pytest.raises(ValueError):
            BoundedGaussianDistribution(name)

        dist = BoundedGaussianDistribution(name, mus=[1.0, 2.0], sigmas=[1.0, 2.0])

        assert dist.nmodes == 2
        assert np.all(np.array(dist["weight"]) == 1.0)

        del dist
        # test log pdf
        dist = BoundedGaussianDistribution(
            name,
            mus=[Uniform(0.0, 1.0, "mu0"), 2.0],
            sigmas=[Uniform(0.0, 1.0, "sigma0"), 2.0],
            weights=DirichletPriorDict(n_dim=2, label="weight"),
        )

        value = 1.0
        hyper = {"mu8": 0.5, "sigma0": 0.5, "weight0": 0.5}
        with pytest.raises(KeyError):
            dist.log_pdf(value, hyper)

        hyper = {"mu0": 0.5, "sigma8": 0.5, "weight0": 0.5}
        with pytest.raises(KeyError):
            dist.log_pdf(value, hyper)

        hyper = {"mu0": 0.5, "sigma0": 0.5, "weight8": 0.5}
        with pytest.raises(KeyError):
            dist.log_pdf(value, hyper)

        hyper = {"mu0": 0.5, "sigma0": 0.5, "weight0": 0.5}
        assert np.isfinite(dist.log_pdf(value, hyper))
        assert np.exp(dist.log_pdf(value, hyper)) == dist.pdf(value, hyper)

        # check negative values give -inf by default
        value = -1.0
        assert dist.log_pdf(value, hyper) == -np.inf

        # check drawn sample is within bounds
        assert dist.low < dist.sample(hyper) < dist.high

        # draw multiple samples
        N = 100
        samples = dist.sample(hyper, size=N)
        assert len(samples) == N
        assert np.all((samples > dist.low) & (samples < dist.high))

        # test with multiple modes
        del dist
        dist = BoundedGaussianDistribution(
            name,
            mus=[Uniform(0.0, 1.0, "mu0"), Uniform(0.0, 1.0, "mu1")],
            sigmas=[Uniform(0.0, 1.0, "sigma0"), Uniform(0.0, 1.0, "sigma1")],
            weights=[0.25, 0.75],
        )

        hyper = {"mu0": 0.5, "sigma0": 0.5, "mu1": 0.7, "sigma1": 0.8}

        N = 100
        samples = dist.sample(hyper, size=N)
        assert len(samples) == N
        assert np.all((samples > dist.low) & (samples < dist.high))

    def test_exponential(self):
        """
        Test the ExponentialDistribution class.
        """

        name = "test"

        dist = ExponentialDistribution(name, mu=1.0)
        assert dist["mu"] == 1.0
        assert dist.fixed["mu"] is True

        dist = ExponentialDistribution(name, mu=Uniform(0.0, 1.0, "mu"))

        value = -1.0
        hyper = {"mu": 0.5}
        assert isinstance(dist["mu"], Uniform)
        assert dist.fixed["mu"] is False
        assert dist.log_pdf(value, hyper) == -np.inf
        assert np.exp(dist.log_pdf(value, hyper)) == dist.pdf(value, hyper)

        # check drawn sample is within bounds
        assert dist.low < dist.sample(hyper) < dist.high

        # draw multiple samples
        N = 100
        samples = dist.sample(hyper, size=N)
        assert len(samples) == N
        assert np.all((samples > dist.low) & (samples < dist.high))

        value = 1.0
        hyper = {"kgsdg": 0.5}
        with pytest.raises(KeyError):
            dist.log_pdf(value, hyper)

    def test_deltafunction(self):
        """
        Test the DeltaFunctionDistribution class.
        """

        name = "test"

        dist = DeltaFunctionDistribution(name, peak=1.0)
        assert dist["peak"] == 1.0
        assert dist.fixed["peak"] is True
        assert np.all(dist.sample(size=10) == 1.0)

        dist = DeltaFunctionDistribution(name, peak=Uniform(0.0, 1.0, "peak"))

        value = 0.1
        hyper = {"peak": 0.5}
        assert isinstance(dist["peak"], Uniform)
        assert dist.fixed["peak"] is False
        assert dist.log_pdf(value, hyper) == -np.inf
        assert np.exp(dist.log_pdf(value, hyper)) == dist.pdf(value, hyper)

        value = 0.5
        assert dist.log_pdf(value, hyper) == 0.0
        assert np.exp(dist.log_pdf(value, hyper)) == dist.pdf(value, hyper)

        # check drawn sample is within bounds
        assert dist.low < dist.sample(hyper) < dist.high

        # draw multiple samples
        N = 100
        samples = dist.sample(hyper, size=N)
        assert len(samples) == N
        assert np.all((samples > dist.low) & (samples < dist.high))

        value = 1.0
        hyper = {"kgsdg": 0.5}
        with pytest.raises(KeyError):
            dist.log_pdf(value, hyper)

    def test_powerlaw(self):
        """
        Test the PowerLawDistribution class.
        """

        name = "test"

        dist = PowerLawDistribution(name, alpha=1.0, minimum=0.1, maximum=10.0)
        assert dist["alpha"] == 1.0
        assert dist.fixed["alpha"] is True
        assert dist["minimum"] == 0.1
        assert dist.fixed["minimum"] is True
        assert dist["maximum"] == 10.0
        assert dist.fixed["maximum"] is True

        # test out of bounds
        with pytest.raises(ValueError):
            PowerLawDistribution(name, alpha=1.0, minimum=-1.0, maximum=10.0)

        with pytest.raises(ValueError):
            PowerLawDistribution(name, alpha=1.0, minimum=-np.inf, maximum=10.0)

        with pytest.raises(ValueError):
            PowerLawDistribution(name, alpha=1.0, minimum=1.0, maximum=0.5)

        with pytest.raises(ValueError):
            PowerLawDistribution(name, alpha=1.0, minimum=1.0, maximum=-np.inf)

        minimum = 0.001
        maximum = 10.0
        dist = PowerLawDistribution(
            name, alpha=Uniform(0.0, 1.0, "alpha"), minimum=minimum, maximum=maximum
        )

        value = -1.0
        hyper = {"alpha": 0.5}
        assert isinstance(dist["alpha"], Uniform)
        assert dist.fixed["alpha"] is False
        assert dist.log_pdf(value, hyper) == -np.inf
        assert np.exp(dist.log_pdf(value, hyper)) == dist.pdf(value, hyper)

        # check drawn sample is within bounds
        assert minimum < dist.sample(hyper) < maximum

        # draw multiple samples
        N = 100
        samples = dist.sample(hyper, size=N)
        assert len(samples) == N
        assert np.all((samples > minimum) & (samples < maximum))

        value = 1.0
        hyper = {"kgsdg": 0.5}
        with pytest.raises(KeyError):
            dist.log_pdf(value, hyper)

    def test_create_distribution(self):
        """
        Test the create_distribution() function.
        """

        name = "test"
        with pytest.raises(ValueError):
            create_distribution(name, "kjbskdvakvkd")

        with pytest.raises(TypeError):
            create_distribution(name, 2.3)

        gausskwargs = {"mus": [1.0, 2.0], "sigmas": [1.0, 2.0]}
        dist = create_distribution(name, "Gaussian", gausskwargs)

        assert isinstance(dist, BoundedGaussianDistribution)
        assert (
            dist["mu0"] == gausskwargs["mus"][0]
            and dist["mu1"] == gausskwargs["mus"][1]
        )
        assert (
            dist["sigma0"] == gausskwargs["sigmas"][0]
            and dist["sigma1"] == gausskwargs["sigmas"][1]
        )
        del dist

        expkwargs = {"mu": Uniform(0.0, 1.0, "mu")}
        dist = create_distribution(name, "Exponential", expkwargs)
        assert isinstance(dist, ExponentialDistribution)
        assert dist["mu"] == expkwargs["mu"]

        newdist = create_distribution(name, dist)
        assert isinstance(newdist, ExponentialDistribution)
        assert newdist["mu"] == dist["mu"]

        deltakwargs = {"peak": Uniform(0.0, 1.0, "peak")}
        dist = create_distribution(name, "DeltaFunction", deltakwargs)
        assert isinstance(dist, DeltaFunctionDistribution)
        assert dist["peak"] == deltakwargs["peak"]

        powerlawkwargs = {
            "alpha": Uniform(-1, 1, name="alpha"),
            "minimum": 0.00001,
            "maximum": 1000.0,
        }
        dist = create_distribution(name, "PowerLaw", powerlawkwargs)
        assert isinstance(dist, PowerLawDistribution)
        assert dist["alpha"] == powerlawkwargs["alpha"]
        assert dist["minimum"] == powerlawkwargs["minimum"]
        assert dist["maximum"] == powerlawkwargs["maximum"]


class TestMassQuadrupoleDistribution(object):
    """
    Test the MassQuadrupoleDistribution object.
    """

    def test_mass_quadrupole_distribution(self):
        # test data sets from bilby
        testdata1 = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "data",
            "hierarchical_test_set_0_result.json",
        )
        testdata2 = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "data",
            "hierarchical_test_set_1_result.json",
        )

        # test invalid gridrange (lower bounds is less than upper bound)
        with pytest.raises(ValueError):
            MassQuadrupoleDistribution(gridrange=[100.0, 1.0])

        # test invalid gridrange (only one value passed)
        with pytest.raises(ValueError):
            MassQuadrupoleDistribution(gridrange=[100.0])

        # test invalid data type
        with pytest.raises(TypeError):
            MassQuadrupoleDistribution(data=1)

        res = ResultList([testdata1, testdata2])

        # remove Q22 from results to test error
        del res[0].posterior["q22"]
        with pytest.raises(RuntimeError):
            MassQuadrupoleDistribution(data=res)

        # distribution with wrong name (i.e., not 'Q22')
        pdist = ExponentialDistribution("Blah", mu=Uniform(0.0, 1e37, "mu"))
        with pytest.raises(ValueError):
            MassQuadrupoleDistribution(data=[testdata1, testdata2], distribution=pdist)

        # distribution with no priors to infer
        pdist = BoundedGaussianDistribution("Q22", mus=[0.0], sigmas=[1e34])
        with pytest.raises(ValueError):
            MassQuadrupoleDistribution(data=[testdata1, testdata2], distribution=pdist)

        # unknown sampler type
        pdist = ExponentialDistribution(
            "Q22", mu=Uniform(name="mu", minimum=0.0, maximum=1e32)
        )
        with pytest.raises(ValueError):
            MassQuadrupoleDistribution(
                data=[testdata1, testdata2], distribution=pdist, sampler="akgkfsfd"
            )

        # unknown bandwidth type for KDE
        bw = "lkgadkgds"
        with pytest.raises(RuntimeError):
            MassQuadrupoleDistribution(
                data=[testdata1, testdata2], distribution=pdist, bw=bw
            ).sample()

        # wrong type for q22grid
        q22grid = "lsgdgkavbc"
        with pytest.raises(TypeError):
            MassQuadrupoleDistribution(
                data=[testdata1, testdata2], distribution=pdist, grid=q22grid
            )

        # test sampler
        mdist = MassQuadrupoleDistribution(
            data=[testdata1, testdata2], distribution=pdist
        )

        # make sure result is None before sampling
        assert mdist.result is None
        with pytest.raises(RuntimeError):
            _ = [_ for _ in mdist.posterior_predictive([0, 1])]

        res = mdist.sample(**{"Nlive": 100, "save": False, "label": "test1"})

        assert isinstance(res, Result)
        assert np.all((res.posterior["mu"] > 0.0) & (res.posterior["mu"] < 1e32))

        # check posterior predictive errors
        with pytest.raises(TypeError):
            _ = [_ for _ in mdist.posterior_predictive(0)]

        # check error when requesting too many samples
        with pytest.raises(ValueError):
            _ = [_ for _ in mdist.posterior_predictive([0, 1], nsamples=10000000000000)]

        points = [1e29, 1e30, 1e31]
        nsamples = 2
        for tfunc in [list, tuple, np.array]:
            assert np.array(
                [
                    values
                    for values in mdist.posterior_predictive(tfunc(points), nsamples)
                ]
            ).shape == (nsamples, len(points))

        del res
        del mdist

        # test using expectation values
        postsamples = ResultList([testdata1, testdata2])

        with pytest.raises(TypeError):
            MassQuadrupoleDistribution(
                data=[testdata1, testdata2],
                distribution=pdist,
                integration_method="expectation",
                nsamples="blah",
            )

        with pytest.raises(ValueError):
            MassQuadrupoleDistribution(
                data=[testdata1, testdata2],
                distribution=pdist,
                integration_method="expectation",
                nsamples=0,
            )

        mdist = MassQuadrupoleDistribution(
            data=[testdata1, testdata2],
            distribution=pdist,
            integration_method="expectation",
        )

        assert len(mdist._posterior_samples[0]) == len(postsamples[0].posterior)
        assert len(mdist._posterior_samples[1]) == len(postsamples[1].posterior)

        del mdist

        # use all samples (but test passing nsamples a larger number)
        mdist = MassQuadrupoleDistribution(
            data=[testdata1, testdata2],
            distribution=pdist,
            integration_method="expectation",
            nsamples=(
                1
                + int(
                    np.max(
                        [len(postsamples[0].posterior), len(postsamples[1].posterior)]
                    )
                )
            ),
        )

        assert len(mdist._posterior_samples[0]) == len(postsamples[0].posterior)
        assert len(mdist._posterior_samples[1]) == len(postsamples[1].posterior)

        del mdist

        # use 500 samples
        nsamples = 500
        mdist = MassQuadrupoleDistribution(
            data=[testdata1, testdata2],
            distribution=pdist,
            integration_method="expectation",
            nsamples=nsamples,
        )

        assert len(mdist._posterior_samples[0]) == nsamples
        assert len(mdist._posterior_samples[1]) == nsamples

        # try sampler
        res = mdist.sample(**{"Nlive": 100, "save": False, "label": "test2"})

        assert isinstance(res, Result)
        assert np.all((res.posterior["mu"] > 0.0) & (res.posterior["mu"] < 1e32))

        del res
        del mdist

        # test grid sampler
        grid = "Blah"
        with pytest.raises(TypeError):
            MassQuadrupoleDistribution(
                data=[testdata1, testdata2], distribution=pdist, grid=grid
            )

        grid = {"mu": 100}
        mdist = MassQuadrupoleDistribution(
            data=[testdata1, testdata2], distribution=pdist, grid=grid
        )

        # make sure result is None before sampling
        assert mdist.result is None

        res = mdist.sample()
        assert isinstance(res, Grid)

        del res
        del mdist

        # test using ellipticity
        pdist = ExponentialDistribution("Q22", mu=Uniform(0.0, 1e-5, "mu"))
        with pytest.raises(ValueError):
            # pdist does not contain ELL
            MassQuadrupoleDistribution(
                data=[testdata1, testdata2], distribution=pdist, use_ellipticity=True
            )

        pdist = ExponentialDistribution("ELL", mu=Uniform(0.0, 1e-5, "mu"))
        mdist = MassQuadrupoleDistribution(
            data=[testdata1, testdata2], distribution=pdist, use_ellipticity=True
        )
        res = mdist.sample(**{"Nlive": 100, "save": False, "label": "test3"})

        assert isinstance(res, Result)
        assert "mu" in res.posterior.columns
        assert np.all((res.posterior["mu"] > 0.0) & (res.posterior["mu"] < 1e-5))

        del res
        del mdist



"""
Test script for simulations.
"""

import glob
import os
import shutil

import bilby
import numpy as np
import pytest
from astropy import units as u
from cwinpy.data import HeterodynedData
from cwinpy.hierarchical import DeltaFunctionDistribution
from cwinpy.pe.simulation import PEPulsarSimulationDAG
from lalpulsar.PulsarParametersWrapper import PulsarParametersPy


class TestPESimulation(object):
    @classmethod
    def setup_class(cls):
        """
        Create directory for simulations.
        """

        # set the base directory
        cls.basedir = os.path.join(
            os.path.split(os.path.realpath(__file__))[0], "simulation"
        )
        os.makedirs(cls.basedir, exist_ok=True)

        # create pulsar parameter files for testing
        cls.pardir = os.path.join(
            os.path.split(os.path.realpath(__file__))[0], "test_pardir"
        )
        os.makedirs(cls.pardir, exist_ok=True)

        cls.names = ["J0000+0000", "J0100+0000"]
        cls.ras = [0.0, (1 / 24) * 2 * np.pi]
        cls.decs = [0.0, 0.0]
        cls.dists = [1.0, None]
        cls.pardict = {}
        for name, ra, dec, dist in zip(cls.names, cls.ras, cls.decs, cls.dists):
            par = PulsarParametersPy()
            par["PSRJ"] = name
            par["F"] = [100.0]  # set frequency to 100 Hz
            par["RAJ"] = ra
            par["DECJ"] = dec
            if dist is not None:
                par["DIST"] = (dist * u.kpc).to("m").value
            with open(os.path.join(cls.pardir, "{}.par".format(name)), "w") as fp:
                fp.write(str(par))
            cls.pardict[name] = os.path.join(cls.pardir, "{}.par".format(name))

        # create heterodyned data for testing
        cls.hetdir = os.path.join(
            os.path.split(os.path.realpath(__file__))[0], "test_hetdir"
        )
        os.makedirs(cls.hetdir, exist_ok=True)

        cls.hetfiles = {}
        for det, name in zip(["H1", "L1"], ["J0000+0000", "J0100+0000"]):
            het = HeterodynedData(
                times=np.linspace(1000000000.0, 1000086340.0, 1440),
                fakeasd=det,
                par=os.path.join(cls.pardir, "{}.par".format(name)),
            )
            cls.hetfiles[det] = os.path.join(cls.hetdir, "{}.hdf".format(name))
            het.write(cls.hetfiles[det], overwrite=True)

    @classmethod
    def teardown_class(cls):
        """
        Remove test simulation directory.
        """

        shutil.rmtree(cls.basedir)
        shutil.rmtree(cls.pardir)
        shutil.rmtree(cls.hetdir)

    def test_failures(self):
        with pytest.raises(TypeError):
            # wrong type for parfile
            PEPulsarSimulationDAG(ampdist=None, parfiles=1.0)

        with pytest.raises(ValueError):
            # wrong type for parfile
            PEPulsarSimulationDAG(ampdist=None, parfiles="blah_blah_blah")

        with pytest.raises(IOError):
            # non-existent par file
            PEPulsarSimulationDAG(ampdist=None, parfiles={"J0000+0000": "no.par"})

        with pytest.raises(ValueError):
            # no amplitude distribution and no parameter files
            PEPulsarSimulationDAG(ampdist=None, parfiles=None)

        with pytest.raises(ValueError):
            # directory clash for pulsar parameter files
            PEPulsarSimulationDAG(
                ampdist=None, parfiles=os.path.join(self.basedir, "pulsars")
            )

        with pytest.raises(TypeError):
            # wrong type for amplitude prior
            PEPulsarSimulationDAG(ampdist=1, parfiles=self.pardir)

        ampprior = bilby.core.prior.Uniform(0.0, 1e40, name="blah")
        with pytest.raises(KeyError):
            # wrong key for amplitude prior
            PEPulsarSimulationDAG(ampdist=ampprior)

        ampprior = bilby.core.prior.Uniform(0.0, 1e40, name="q22")

        with pytest.raises(TypeError):
            # wrong type for distance error
            PEPulsarSimulationDAG(ampdist=ampprior, distance_err=1)

        with pytest.raises(TypeError):
            # wrong type for prior
            PEPulsarSimulationDAG(ampdist=ampprior, prior=1)

        with pytest.raises(ValueError):
            # empty dictionary for prior
            PEPulsarSimulationDAG(ampdist=ampprior, prior={})

        with pytest.raises(FileNotFoundError):
            # bad prior file name
            PEPulsarSimulationDAG(ampdist=ampprior, prior="ksdkfkhvsad")

        with pytest.raises(TypeError):
            # wrong type for number of pulsars
            PEPulsarSimulationDAG(ampdist=ampprior, npulsars=2.3)

        with pytest.raises(ValueError):
            # wrong number of pulsars
            PEPulsarSimulationDAG(ampdist=ampprior, npulsars=0)

        with pytest.raises(TypeError):
            # wrong type for position distribution
            PEPulsarSimulationDAG(ampdist=ampprior, posdist=1, npulsars=1)

        with pytest.raises(KeyError):
            # wrong key for position distribution
            PEPulsarSimulationDAG(
                ampdist=ampprior,
                parfiles=self.pardir,
                posdist=bilby.core.prior.PriorDict(
                    {"blah": bilby.core.prior.Uniform(1, 2, name="blah")}
                ),
            )

        with pytest.raises(ValueError):
            # position distribution with unknown values
            PEPulsarSimulationDAG(
                ampdist=ampprior,
                npulsars=1,
                posdist=bilby.core.prior.PriorDict(
                    {"blah": bilby.core.prior.Uniform(1, 2, name="blah")}
                ),
            )

        with pytest.raises(TypeError):
            # wrong type for frequency distribution
            PEPulsarSimulationDAG(ampdist=ampprior, fdist=1, npulsars=1)

        with pytest.raises(TypeError):
            # wrong type for frequency distribution
            PEPulsarSimulationDAG(ampdist=ampprior, fdist=1, npulsars=1)

        with pytest.raises(TypeError):
            # wrong type for orientation distribution
            PEPulsarSimulationDAG(ampdist=ampprior, oridist=1, npulsars=1)

        with pytest.raises(TypeError):
            # wrong type for detector
            PEPulsarSimulationDAG(ampdist=ampprior, npulsars=1, detector=1)

        with pytest.raises(ValueError):
            # no detectors or data files
            PEPulsarSimulationDAG(
                ampdist=ampprior,
                npulsars=1,
                basedir=os.path.join(self.basedir, "nodet"),
                detector=None,
            )

        shutil.rmtree(os.path.join(self.basedir, "nodet"))

    def test_sim_pulsars(self):
        """
        Test the code for when generating fake pulsars.
        """

        # use delta function priors to check values are the same
        q22 = 1.1e32
        ra = 1.0
        dec = 0.5
        dist = 2.5
        iota = 0.1
        phi0 = 2.3
        psi = 0.5
        f0 = 100.0
        ampprior = bilby.core.prior.DeltaFunction(q22, name="q22")
        posdist = bilby.core.prior.PriorDict(
            {
                "ra": bilby.core.prior.DeltaFunction(ra, name="ra"),
                "dec": bilby.core.prior.DeltaFunction(dec, name="dec"),
                "dist": bilby.core.prior.DeltaFunction(dist, name="dist"),
            }
        )
        oridist = bilby.core.prior.PriorDict(
            {
                "psi": bilby.core.prior.DeltaFunction(psi, name="psi"),
                "phi0": bilby.core.prior.DeltaFunction(phi0, name="phi0"),
                "iota": bilby.core.prior.DeltaFunction(iota, name="iota"),
            }
        )
        fdist = bilby.core.prior.DeltaFunction(f0, name="f0")

        # the expected default PE prior
        expectedprior = bilby.core.prior.PriorDict(
            {
                "h0": bilby.core.prior.Uniform(
                    0.0, 1e-22, name="h0", latex_label="$h_0$"
                ),
                "iota": bilby.core.prior.Sine(name="iota", latex_label=r"$\iota$"),
                "phi0": bilby.core.prior.Uniform(
                    0.0, np.pi, name="phi0", latex_label=r"$\phi_0$"
                ),
                "psi": bilby.core.prior.Uniform(
                    0.0, np.pi / 2, name="psi", latex_label=r"$\psi$"
                ),
            }
        )

        npulsars = 2
        detectors = "H1"
        testdir = os.path.join(self.basedir, "test_sim_pulsar")
        sim = PEPulsarSimulationDAG(
            ampdist=ampprior,
            oridist=oridist,
            posdist=posdist,
            npulsars=npulsars,
            detector=detectors,
            basedir=testdir,
            fdist=fdist,
        )

        # test error for start/end time setting
        with pytest.raises(ValueError):
            PEPulsarSimulationDAG(
                ampdist=ampprior,
                oridist=oridist,
                posdist=posdist,
                npulsars=npulsars,
                detector=detectors,
                basedir=testdir,
                fdist=fdist,
                starttime=1020304050,
            )

        shutil.rmtree(testdir)

        sim = PEPulsarSimulationDAG(
            ampdist=ampprior,
            oridist=oridist,
            posdist=posdist,
            npulsars=npulsars,
            detector=detectors,
            basedir=testdir,
            fdist=fdist,
        )

        # check directories
        for dir in glob.glob(os.path.join(testdir, "*")):
            assert os.path.basename(dir) in [
                "configs",
                "error",
                "log",
                "out",
                "priors",
                "pulsars",
                "results",
                "submit",
            ]

        for dir in ["configs", "priors", "pulsars"]:
            assert len(glob.glob(os.path.join(testdir, dir, "*"))) == npulsars

        # check priors are the same as the default
        for pname in sim.priors:
            assert sim.priors[pname] == expectedprior

            # check fake pulsars contain the same values
            psr = PulsarParametersPy(
                os.path.join(testdir, "pulsars", "{}.par".format(pname))
            )

            assert psr["PSRJ"] == pname
            assert psr["RAJ"] == ra
            assert psr["DECJ"] == dec
            assert psr["F0"] == f0
            assert psr["Q22"] == q22
            assert psr["PSI"] == psi
            assert psr["IOTA"] == iota
            assert psr["PHI0"] == phi0
            assert np.allclose(psr["DIST"], (dist * u.kpc).to("m").value)

        shutil.rmtree(testdir)

        # test galactic and galactocentric position distributions
        l = 0.3  # noqa: E741
        b = 1.1
        posdist = bilby.core.prior.PriorDict(
            {
                "l": bilby.core.prior.DeltaFunction(l, name="l"),
                "b": bilby.core.prior.DeltaFunction(b, name="b"),
                "dist": bilby.core.prior.DeltaFunction(dist, name="dist"),
            }
        )

        sim = PEPulsarSimulationDAG(
            ampdist=ampprior,
            oridist=oridist,
            posdist=posdist,
            npulsars=npulsars,
            detector=detectors,
            basedir=testdir,
            fdist=fdist,
        )

        ras = []
        decs = []
        dists = []
        for pname in sim.priors:
            # check fake pulsars contain the same values
            psr = PulsarParametersPy(
                os.path.join(testdir, "pulsars", "{}.par".format(pname))
            )

            ras.append(psr["RAJ"])
            decs.append(psr["DECJ"])
            dists.append(psr["DIST"])

        for vals in [ras, decs, dists]:
            assert np.all(np.array(vals) == vals[0])

        shutil.rmtree(testdir)

        x = 0.3
        y = 1.1
        z = 0.2
        posdist = bilby.core.prior.PriorDict(
            {
                "x": bilby.core.prior.DeltaFunction(x, name="x"),
                "y": bilby.core.prior.DeltaFunction(y, name="y"),
                "z": bilby.core.prior.DeltaFunction(z, name="z"),
            }
        )

        sim = PEPulsarSimulationDAG(
            ampdist=ampprior,
            oridist=oridist,
            posdist=posdist,
            npulsars=npulsars,
            detector=detectors,
            basedir=testdir,
            fdist=fdist,
        )

        ras = []
        decs = []
        dists = []
        for pname in sim.priors:
            # check fake pulsars contain the same values
            psr = PulsarParametersPy(
                os.path.join(testdir, "pulsars", "{}.par".format(pname))
            )

            ras.append(psr["RAJ"])
            decs.append(psr["DECJ"])
            dists.append(psr["DIST"])

        for vals in [ras, decs, dists]:
            assert np.all(np.array(vals) == vals[0])

        shutil.rmtree(testdir)

    def test_data_pulsars(self):
        """
        Test the code for when using existing pulsar data.
        """

        # use delta function priors to check values are the same
        q22 = 1.1e32
        iota = 0.1
        phi0 = 2.3
        psi = 0.5
        f0 = 100.0
        dist = 2.0
        ampprior = bilby.core.prior.DeltaFunction(q22, name="q22")
        oridist = bilby.core.prior.PriorDict(
            {
                "psi": bilby.core.prior.DeltaFunction(psi, name="psi"),
                "phi0": bilby.core.prior.DeltaFunction(phi0, name="phi0"),
                "iota": bilby.core.prior.DeltaFunction(iota, name="iota"),
            }
        )
        posdist = bilby.core.prior.PriorDict(
            {"dist": bilby.core.prior.DeltaFunction(dist, name="dist")}
        )
        fdist = bilby.core.prior.DeltaFunction(f0, name="f0")

        testdir = os.path.join(self.basedir, "test_data_pulsar")

        # set different priors for the pulsars
        priors = {}
        disterrs = {}
        for i, name in enumerate(self.names):
            priors[name] = bilby.core.prior.PriorDict(
                {"q22": bilby.core.prior.Uniform(0.0, (i + 1) * 1e40, name="q22")}
            )
            disterrs[name] = (i + 1) * 0.2

        # pass directory of par files and data files
        sim = PEPulsarSimulationDAG(
            ampdist=ampprior,
            prior=priors,
            distance_err=disterrs,
            oridist=oridist,
            posdist=posdist,
            parfiles=self.pardir,
            datafiles=self.hetfiles,
            basedir=testdir,
            fdist=fdist,
            overwrite_parameters=False,
        )

        for dir in glob.glob(os.path.join(testdir, "*")):
            assert os.path.basename(dir) in [
                "configs",
                "error",
                "log",
                "out",
                "priors",
                "pulsars",
                "results",
                "submit",
            ]

        for dir in ["configs", "priors", "pulsars"]:
            assert len(glob.glob(os.path.join(testdir, dir, "*"))) == len(self.ras)

        # check signal values are correct
        for i, pname in enumerate(self.names):
            # check fake pulsars contain the same values
            psr = PulsarParametersPy(
                os.path.join(testdir, "pulsars", "{}.par".format(pname))
            )

            assert psr["PSRJ"] == pname
            assert psr["RAJ"] == self.ras[i]
            assert psr["DECJ"] == self.decs[i]
            assert psr["F0"] == f0
            assert psr["Q22"] == q22
            assert psr["PSI"] == psi
            assert psr["IOTA"] == iota
            assert psr["PHI0"] == phi0

            if self.dists[i] is None:
                assert np.allclose(psr["DIST"], (dist * u.kpc).to("m").value)
            else:
                assert np.allclose(psr["DIST"], (self.dists[i] * u.kpc).to("m").value)

            # check the priors
            assert sim.priors[pname]["q22"] == priors[pname]["q22"]
            assert "dist" in sim.priors[pname]
            if self.dists[i] is None:
                assert np.allclose(
                    sim.priors[pname]["dist"].sigma,
                    (dist * (i + 1) * 0.2 * u.kpc).to("m").value,
                )
                assert np.allclose(
                    sim.priors[pname]["dist"].mu, (dist * u.kpc).to("m").value
                )
            else:
                assert np.allclose(
                    sim.priors[pname]["dist"].sigma,
                    (self.dists[i] * (i + 1) * 0.2 * u.kpc).to("m").value,
                )
                assert np.allclose(
                    sim.priors[pname]["dist"].mu, (self.dists[i] * u.kpc).to("m").value
                )
            assert sim.priors[pname]["dist"].minimum == 0.0
            assert not np.isfinite(sim.priors[pname]["dist"].maximum)

        shutil.rmtree(testdir)

        # now add distance errors with a single float and with individual distributions
        disterrs = 0.3
        sim = PEPulsarSimulationDAG(
            ampdist=ampprior,
            prior=priors,
            distance_err=disterrs,
            oridist=oridist,
            posdist=posdist,
            parfiles=self.pardict,  # pass dictionary or par files this time
            datafiles=self.hetfiles,
            basedir=testdir,
            fdist=fdist,
            overwrite_parameters=True,
        )

        # check signal values are correct
        for i, pname in enumerate(self.names):
            # check fake pulsars contain the same values
            psr = PulsarParametersPy(
                os.path.join(testdir, "pulsars", "{}.par".format(pname))
            )

            assert psr["PSRJ"] == pname
            assert psr["RAJ"] == self.ras[i]
            assert psr["DECJ"] == self.decs[i]
            assert psr["F0"] == f0
            assert psr["Q22"] == q22
            assert psr["PSI"] == psi
            assert psr["IOTA"] == iota
            assert psr["PHI0"] == phi0

            # both distances should be the same as ones in the par file will be
            # overwritten
            assert np.allclose(psr["DIST"], (dist * u.kpc).to("m").value)

            # check the priors
            assert sim.priors[pname]["q22"] == priors[pname]["q22"]
            assert "dist" in sim.priors[pname]
            assert np.allclose(
                sim.priors[pname]["dist"].sigma,
                (dist * disterrs * u.kpc).to("m").value,
            )
            assert np.allclose(
                sim.priors[pname]["dist"].mu, (dist * u.kpc).to("m").value
            )
            assert sim.priors[pname]["dist"].minimum == 0.0
            assert not np.isfinite(sim.priors[pname]["dist"].maximum)

        shutil.rmtree(testdir)

        disterrs = {
            self.names[0]: bilby.core.prior.Uniform(
                (self.dists[0] * 0.9 * u.kpc).to("m").value,
                (self.dists[0] * 1.1 * u.kpc).to("m").value,
                name="dist",
            )
        }
        sim = PEPulsarSimulationDAG(
            ampdist=ampprior,
            prior=priors,
            distance_err=disterrs,
            oridist=oridist,
            posdist=posdist,
            parfiles=self.pardict,  # pass dictionary or par files this time
            datafiles=self.hetfiles,
            basedir=testdir,
            fdist=fdist,
            sampler_kwargs={"nlive": 2000},
            overwrite_parameters=False,
        )

        # check signal values are correct
        for i, pname in enumerate(self.names):
            # check fake pulsars contain the same values
            psr = PulsarParametersPy(
                os.path.join(testdir, "pulsars", "{}.par".format(pname))
            )

            assert psr["PSRJ"] == pname
            assert psr["RAJ"] == self.ras[i]
            assert psr["DECJ"] == self.decs[i]
            assert psr["F0"] == f0
            assert psr["Q22"] == q22
            assert psr["PSI"] == psi
            assert psr["IOTA"] == iota
            assert psr["PHI0"] == phi0

            if self.dists[i] is None:
                assert np.allclose(psr["DIST"], (dist * u.kpc).to("m").value)
            else:
                assert np.allclose(psr["DIST"], (self.dists[i] * u.kpc).to("m").value)

            # check the priors
            assert sim.priors[pname]["q22"] == priors[pname]["q22"]

            if self.dists[i] is not None:
                assert "dist" in sim.priors[pname]
                assert sim.priors[pname]["dist"] == disterrs[pname]
            else:
                assert "dist" not in sim.priors[pname]

        shutil.rmtree(testdir)

        # use a single prior for every pulsar
        priors = bilby.core.prior.PriorDict(
            {"h0": bilby.core.prior.Uniform(0.0, 1e-22, name="h0")}
        )

        # use h0 to set amplitude
        h0 = 1e-24
        ampprior = bilby.core.prior.DeltaFunction(h0, name="h0")

        # pass directory of par files and data files
        sim = PEPulsarSimulationDAG(
            ampdist=ampprior,
            prior=priors,
            oridist={},  # set default orientation
            posdist=posdist,
            parfiles=self.pardir,
            datafiles=self.hetfiles,
            basedir=testdir,
            fdist=fdist,
        )

        # check signal values are correct
        for i, pname in enumerate(self.names):
            # check fake pulsars contain the same values
            psr = PulsarParametersPy(
                os.path.join(testdir, "pulsars", "{}.par".format(pname))
            )

            assert psr["H0"] == h0
            assert 0.0 <= psr["PSI"] <= np.pi / 2.0
            assert 0.0 <= psr["IOTA"] <= np.pi
            assert 0.0 <= psr["PHI0"] <= np.pi

            # check the priors
            assert sim.priors[pname] == priors

        shutil.rmtree(testdir)

        # use a single prior file for each pulsar
        os.makedirs(os.path.join(testdir, "test_prior"), exist_ok=True)
        priors = bilby.core.prior.PriorDict(
            {"q22": bilby.core.prior.Uniform(0.0, 1e40, name="q22")}
        )
        priors.to_file(outdir=os.path.join(testdir, "test_prior"), label="test")
        priorfile = os.path.join(testdir, "test_prior", "test.prior")

        # use ellipticity to set amplitude distribution
        epsilon = 1e-7
        ampprior = DeltaFunctionDistribution(name="epsilon", peak=epsilon)

        # pass directory of par files and data files
        sim = PEPulsarSimulationDAG(
            ampdist=ampprior,
            prior=priorfile,
            oridist={},  # set default orientation
            posdist=posdist,
            parfiles=self.pardir,
            datafiles=self.hetfiles,
            basedir=testdir,
            fdist=fdist,
        )

        # check signal values are correct
        for i, pname in enumerate(self.names):
            # check fake pulsars contain the same values
            psr = PulsarParametersPy(
                os.path.join(testdir, "pulsars", "{}.par".format(pname))
            )

            assert psr["Q22"] == epsilon * 1e38 * np.sqrt(15 / (8 * np.pi))
            assert 0.0 <= psr["PSI"] <= np.pi / 2.0
            assert 0.0 <= psr["IOTA"] <= np.pi
            assert 0.0 <= psr["PHI0"] <= np.pi

            # check the priors
            assert sim.priors[pname] == priors

        shutil.rmtree(testdir)



# -*- coding: utf-8 -*-
#
# cwinpy documentation build configuration file (copied from that for bilby)
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import subprocess
import sys

import cwinpy

sys.path.insert(0, os.path.abspath("../cwinpy/"))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "sphinx.ext.intersphinx",
    "numpydoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.autosectionlabel",
    "sphinx_tabs.tabs",
    "recommonmark",
    "sphinx.ext.todo",
    "sphinxcontrib.images",
]

autosummary_generate = True
autodoc_docstring_signature = True
autodoc_default_options = {"members-order": "groupwise", "show-inheritance": True}
todo_include_todos = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = {
    ".rst": "restructuredtext",
    ".txt": "restructuredtext",
    ".md": "markdown",
}

# The master toctree document.
master_doc = "index"

# General information about the project.
project = u"cwinpy"
copyright = u"2019, Matthew Pitkin"
author = u"Matthew Pitkin"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
fullversion = cwinpy.__version__

# The short X.Y version.
version = ".".join(fullversion.split(".")[:2])

# The full version, including alpha/beta/rc tags.
release = fullversion

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "requirements.txt",
    "comparisons/scripts/outputs/*_data.txt",
    "comparisons/scripts/outputs/*_prior.txt",
    "comparisons/scripts/outputs/*_SNR",
    "comparisons/scripts/outputs/*_Znoise",
    "comparisons/scripts/outputs/*.par",
    "comparisons/scripts/outputs/*.hdf",
    "comparisons/scripts/outputs/*.json",
    "comparisons/scripts/outputs/*.json.old",
    "comparisons/scripts/data",
    "pe/pe_help.txt",
    "pe/cwinpy_pe_dag.ini",
    "pe/data/*.txt",
    "pe/data/*.par",
    "pe/data/*.ini",
    "pe/data/*.txt.gz",
    "pe/data/example1/*.json",
    "pe/data/example1/*.pickle",
    "pe/data/example1/*_trace.png",
    "pe/data/example2/*.json",
    "pe/data/example2/*.pickle",
    "pe/data/example2/*_trace.png",
]

add_function_parentheses = True
add_module_names = False
show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    "**": [
        "about.html",
        "navigation.html",
        "relations.html",  # needs 'show_related': True theme option to display
        "searchbox.html",
        "donate.html",
    ]
}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "cwinpydoc"

# add logo
html_logo = "logo.png"
html_favicon = "logo_cw.svg"

html_theme_options = {
    "style_nav_header_background": "#e3e2e1",
    "logo_only": True,
}

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "cwinpy.tex", u"cwinpy Documentation", u"Matthew Pitkin", "manual"),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "cwinpy", u"cwinpy Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "cwinpy",
        u"cwinpy Documentation",
        author,
        "cwinpy",
        "One line description of project.",
        "Miscellaneous",
    ),
]

numpydoc_show_class_members = False

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://docs.scipy.org/doc/numpy/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/reference/", None),
    "matplotlib": ("https://matplotlib.org/", None),
    "astropy": ("http://docs.astropy.org/en/stable/", None),
    "bilby": ("https://lscsoft.docs.ligo.org/bilby/", None),
    "pycondor": ("https://jrbourbeau.github.io/pycondor/", None),
    "numba": ("http://numba.pydata.org/numba-doc/latest/", None),
    "gwpy": ("https://gwpy.github.io/docs/stable/", None),
}


# generate cwinpy_pe help string for docs
pe_help = subprocess.getoutput("cwinpy_pe --help")
with open("pe/pe_help.txt", "w") as fp:
    fp.write(pe_help)



"""
A selection of general utility functions.
"""

import ctypes
import os
import string
import subprocess
from functools import reduce
from math import gcd

import lalpulsar
import numpy as np
from astropy import units as u
from lalpulsar.PulsarParametersWrapper import PulsarParametersPy
from numba import jit, njit
from numba.extending import get_cython_function_address

# create a numba-ified version of scipy's gammaln function (see, e.g.
# https://github.com/numba/numba/issues/3086#issuecomment-403469308)
addr = get_cython_function_address("scipy.special.cython_special", "gammaln")
functype = ctypes.CFUNCTYPE(ctypes.c_double, ctypes.c_double, ctypes.c_int)
gammaln_fn = functype(addr)


@njit
def gammaln(x):
    return gammaln_fn(x, 0)  # 0 is for the required int!


@jit(nopython=True)
def logfactorial(n):
    """
    The natural logarithm of the factorial of an integer using the fact that

    .. math::

        \\ln{(n!)} = \\ln{\\left(\\Gamma (n+1)\\right)}

    Parameters
    ----------
    n: int
        An integer for which the natural logarithm of its factorial is
        required.

    Returns
    -------
    float
    """

    return gammaln(n + 1)


def gcd_array(denominators):
    """
    Function to calculate the greatest common divisor of a list of values.
    """

    if not isinstance(denominators, (tuple, list, np.ndarray)):
        raise TypeError("Must have a list or array")

    denoms = np.asarray(denominators).flatten()  # 1d array
    denoms = denoms.astype(np.int)

    if len(denoms) < 2:
        raise ValueError("Must have more than two values")

    return reduce(lambda a, b: gcd(a, b), denoms)


def is_par_file(parfile):
    """
    Check if a TEMPO-style pulsar parameter file is valid.

    Parameters
    ----------
    parfile: str
        The path to a file.

    Returns
    -------
    ispar: bool
        Returns True is if it is a valid parameter file.
    """

    # check that the file is ASCII text (see, e.g.,
    # https://stackoverflow.com/a/1446571/1862861)

    if os.path.isfile(parfile):
        msg = subprocess.Popen(
            ["file", "--mime", "--dereference", parfile], stdout=subprocess.PIPE
        ).communicate()[0]
        if "text/plain" in msg.decode():
            psr = PulsarParametersPy(parfile)
            # file must contain a frequency, right ascension, declination and
            # pulsar name
            # file must contain right ascension and declination
            if (
                psr["F"] is None
                or (psr["RAJ"] is None and psr["RA"] is None)
                or (psr["DECJ"] is None and psr["DEC"] is None)
                or get_psr_name(psr) is None
            ):
                return False

            return True

    return False


def get_psr_name(psr):
    """
    Get the pulsar name from the TEMPO(2)-style parameter files by trying the
    keys "PSRJ", "PSRB", "PSR", and "NAME" in that order of precedence.

    Parameters
    ----------
    psr: PulsarParameterPy
        A PulsarParameterPy object

    Returns
    -------
    name: str
        The string containing the name or None if not found.
    """

    for name in ["PSRJ", "PSRB", "PSR", "NAME"]:
        if psr[name] is not None:
            return psr[name]

    return None


def int_to_alpha(pos, case="upper"):
    """
    For an integer number generate a corresponding alphabetical string
    following the mapping: 1 -> "A", 2 -> "B", ..., 27 -> "AA", 28 -> "AB", ...

    Parameters
    ----------
    pos: int
        A positive integer greater than 0.
    case: str
        Set whether to use upper or lower case. Default is "upper".

    Returns
    -------
    alpha: str
        The alphabetical equivalent
    """

    if case == "lower":
        alphas = list(string.ascii_lowercase)
    else:
        alphas = list(string.ascii_uppercase)

    if not isinstance(pos, int):
        raise TypeError("Value must be an integer")

    if pos < 1:
        raise ValueError("Value must be greater than zero")

    count = pos - 1
    result = ""
    while count >= 0:
        result = alphas[int(count) % len(alphas)] + result
        count /= len(alphas)
        count -= 1

    return result


def ellipticity_to_q22(epsilon, units=False):
    """
    Convert the fiducial ellipticity :math:`\\varepsilon` to the mass
    quadrupole :math:`Q_{22}` (in units of kg m\ :sup:`2`) via (see, e.g.,
    Equation A3 of [1]_):

    .. math::

        Q_{22} = \\varepsilon I_{38} \\sqrt{\\frac{15}{8\\pi}},

    where :math:`I_{38} = 10^{38}` kg m\ :sup:`2` is the canonical moment of
    inertia of the star.

    References
    ----------

    .. [1] B. Abbott, et al., Ap. J., 879, 10, 2019
       (`arXiv:1902.08507 <https://arxiv.org/abs/1902.08507>`_)

    Parameters
    ----------
    epsilon: float, array_like
        A value, or array of values, of the mass quadrupole.
    units: bool
        If True add units to the output.

    Returns
    -------
    q22: float, array_like
        A value, or array of values, of the mass quadrupole.
    """

    if isinstance(epsilon, float):
        q22 = epsilon * 1e38 * np.sqrt(15.0 / (8.0 * np.pi))
    else:
        q22 = np.array(epsilon) * 1e38 * np.sqrt(15.0 / (8.0 * np.pi))

    if units:
        # add units
        return q22 * u.kg * u.m ** 2
    else:
        return q22


def q22_to_ellipticity(q22):
    """
    Convert mass quadrupole :math:`Q_{22}` (in units of kg m\ :sup:`2`) to
    fidicual ellipticity via (see, e.g., Equation A3 of [1]_):

    .. math::

        \\varepsilon = \\frac{Q_{22}}{I_{38}}\\sqrt{\\frac{8\\pi}{15}},

    where :math:`I_{38} = 10^{38}` kg m\ :sup:`2` is the canonical moment of
    inertia of the star.

    References
    ----------

    .. [1] B. Abbott, et al., Ap. J., 879, 10, 2019
       (`arXiv:1902.08507 <https://arxiv.org/abs/1902.08507>`_)

    Parameters
    ----------
    q22: float, array_like
        A value, or array of values, of the mass quadrupole.

    Returns
    -------
    ellipticity: float, array_like
        A value, or array of values, of the fiducial ellipticity.
    """

    if isinstance(q22, list):
        ellipticity = (np.array(q22) / 1e38) * np.sqrt(8.0 * np.pi / 15.0)
    else:
        ellipticity = (q22 / 1e38) * np.sqrt(8.0 * np.pi / 15.0)

    if hasattr(q22, "value"):
        # remove units if value/array has astropy units
        return ellipticity.value
    else:
        return ellipticity


def initialise_ephemeris(
    ephem="DE405", units="TCB", earthfile=None, sunfile=None, timefile=None
):
    """
    Download/read and return solar system ephemeris and time coordinate data.
    If files are provided these will be used and read. If not provided then,
    using supplied ``ephem`` and ``units`` values, it will first attempt to
    find files locally (either in your current path or in a path supplied by
    a ``LAL_DATA_PATH`` environment variable), and if not present will then
    attempt to download the files from a repository.

    To do
    -----

    Add the ability to create ephemeris files using astropy.

    Parameters
    ----------
    earthfile: str
        A file containing the Earth's position/velocity ephemeris
    sunfile: str
        A file containing the Sun's position/velocity ephemeris
    timefile: str
        A file containing time corrections for the TCB or TDB time coordinates.
    ephem: str
        The JPL ephemeris name, e.g., DE405
    units: str
        The time coordinate system, which can be either "TDB" or "TCB" (TCB is
        the default).
    """

    DOWNLOAD_URL = "https://git.ligo.org/lscsoft/lalsuite/raw/master/lalpulsar/lib/{}"

    unit = None
    if timefile is None:
        if units.upper() in ["TCB", "TDB"]:
            unit = dict(TCB="te405", TDB="tdb")[units.upper()]
        else:
            raise ValueError("units must be TCB or TDB")

    earth = "earth00-40-{}.dat.gz".format(ephem) if earthfile is None else earthfile
    sun = "sun00-40-{}.dat.gz".format(ephem) if sunfile is None else sunfile
    time = "{}_2000-2040.dat.gz".format(unit) if timefile is None else timefile

    try:
        edat = lalpulsar.InitBarycenter(earth, sun)
    except RuntimeError:
        # try downloading the ephemeris files
        try:
            from astropy.utils.data import download_file

            efile = download_file(DOWNLOAD_URL.format(earth), cache=True)
            sfile = download_file(DOWNLOAD_URL.format(sun), cache=True)
            edat = lalpulsar.InitBarycenter(efile, sfile)
        except Exception as e:
            raise IOError("Could not read in ephemeris files: {}".format(e))

    try:
        tdat = lalpulsar.InitTimeCorrections(time)
    except RuntimeError:
        try:
            # try downloading the time coordinate file
            from astropy.utils.data import download_file

            tfile = download_file(DOWNLOAD_URL.format(time), cache=True)
            tdat = lalpulsar.InitTimeCorrections(tfile)
        except Exception as e:
            raise IOError("Could not read in time correction file: {}".format(e))

    return edat, tdat



"""
Test script for testing class.
"""

import os
import shutil

import bilby
import numpy as np
import pycondor
import pytest
from cwinpy.pe.testing import PEPPPlotsDAG


class TestPEPP(object):
    @classmethod
    def setup_class(cls):
        """
        Create directory for tests and set default values.
        """

        # set the base directory
        cls.basedir = os.path.join(os.path.split(os.path.realpath(__file__))[0], "base")

        cls.ninj = 50  # number of simulated signals
        cls.maxamp = 5e-23  # maximum amplitude
        cls.freqrange = (10.0, 100.0)  # frequency range

        # default prior dictionary
        cls.priors = {}
        cls.priors["h0"] = bilby.core.prior.Uniform(
            name="h0", minimum=0.0, maximum=1e-22
        )

    @classmethod
    def teardown_class(cls):
        """
        Remove test directory.
        """

        shutil.rmtree(cls.basedir)

    def test_failures(self):
        with pytest.raises(TypeError):
            PEPPPlotsDAG(1)

        with pytest.raises(ValueError):
            PEPPPlotsDAG(self.priors, ninj=-1)

        with pytest.raises(ValueError):
            PEPPPlotsDAG(self.priors, maxamp=-1.0)

        with pytest.raises(IOError):
            PEPPPlotsDAG(self.priors, basedir=1)

        with pytest.raises(TypeError):
            PEPPPlotsDAG(self.priors, basedir=self.basedir, freqrange=1)

        with pytest.raises(ValueError):
            PEPPPlotsDAG(self.priors, basedir=self.basedir, freqrange=[1, 2, 3])

    def test_run(self):
        run = PEPPPlotsDAG(
            self.priors,
            basedir=self.basedir,
            ninj=self.ninj,
            maxamp=self.maxamp,
            freqrange=self.freqrange,
        )

        assert len(run.pulsars) == self.ninj
        assert np.all(
            np.array([run.pulsars[psr]["parameters"]["H0"] for psr in run.pulsars])
            < self.maxamp
        )
        assert np.all(
            np.array([run.pulsars[psr]["parameters"]["F0"] for psr in run.pulsars])
            > self.freqrange[0]
        ) and np.all(
            np.array([run.pulsars[psr]["parameters"]["F0"] for psr in run.pulsars])
            < self.freqrange[1]
        )

        # check output prior
        for prior1, prior2 in zip(
            bilby.core.prior.PriorDict(filename=run.priorfile),
            bilby.core.prior.PriorDict(dictionary=self.priors),
        ):
            assert prior1 == prior2

        # check for the correct number of pulsars
        assert len(run.pulsars) == self.ninj

        # check for the correct number of output parameter files
        assert len(os.listdir(run.pulsardir)) == self.ninj

        # check output is a DAG
        assert isinstance(run.runner.dag.pycondor_dag, pycondor.Dagman)

        # checkout correct number of DAG jobs
        assert len(run.runner.dag.pycondor_dag.nodes) == (self.ninj + 1)

        # check config files are present
        configexists = 0
        for psr in run.pulsars:
            for job in run.runner.dag.pycondor_dag.nodes[:-1]:
                if psr in str(job.args):
                    if os.path.isfile(
                        os.path.join(
                            self.basedir, "configs", job.args[0].arg.split()[-1]
                        )
                    ):
                        configexists += 1
                        break

        assert configexists == self.ninj



"""
Classes providing likelihood functions.
"""

import re

import bilby
import lal
import numpy as np
from lalpulsar.PulsarParametersWrapper import PulsarParametersPy
from lalpulsar.simulateHeterodynedCW import HeterodynedCWSimulator
from numba import jit, types
from numba.typed import Dict as numbadict

from .data import HeterodynedData, MultiHeterodynedData
from .utils import logfactorial


class TargetedPulsarLikelihood(bilby.core.likelihood.Likelihood):
    """
    A likelihood, based on the :class:`bilby.core.likelihood.Likelihood`, for
    use in source parameter estimation for continuous wave signals, with
    particular focus on emission from known pulsars. As a default, this class
    will assume a Student's t-likelihood function, as defined in Equation
    12 of [1]_. This likelihood function assumes that the noise process in each
    dataset can be broken down into multiple chunks of zero mean stationary
    Gaussian noise, each with unknown, and independent, values of the standard
    deviation (see :meth:`~cwinpy.data.HeterodynedData.bayesian_blocks` and
    Appendix B of [1]_). The Gaussian likelihood for each chunk is analytically
    marginalised over the standard deviation, using a Jeffreys prior, and the
    product of the likelihoods for each gives the overall likelihood function.
    A Gaussian likelihood function can also be used (Equation 13 of [1]_), if
    estimates of the noise standard deviation for each chunk are available.

    Parameters
    ----------
    data: (str, HeterodynedData, MultiHeterodynedData)
        A :class:`~cwinpy.data.HeterodynedData` or
        :class:`~cwinpy.data.MultiHeterodynedData` object containing all
        required data.
    priors: :class:`bilby.core.prior.PriorDict`
        A :class:`bilby.core.prior.PriorDict` containing the parameters that
        are to be sampled. This is required to be set before this class is
        initialised.
    likelihood: str
        A string setting which likelihood function to use. This can either be
        'studentst' (the default), or 'gaussian' ('roq' should be added in the
        future).
    numba: bool
        Boolean to set whether to use the `numba` JIT compiled version of the
        likelihood function.

    References
    ----------

    .. [1] M. Pitkin, M. Isi, J. Veitch & G. Woan, `arXiv:1705.08978v1
       <https:arxiv.org/abs/1705.08978v1>`_, 2017.
    """

    # a set of parameters that define the "amplitude" of the signal (i.e.,
    # they do not effect the phase evolution of the signal)
    AMPLITUDE_PARAMS = [
        "H0",
        "PHI0",
        "PSI",
        "IOTA",
        "COSIOTA",
        "C21",
        "C22",
        "PHI21",
        "PHI22",
        "I21",
        "I31",
        "LAMBDA",
        "COSTHETA",
        "THETA",
        "Q22",
        "DIST",
    ]

    # amplitude parameters for the "source" model
    SOURCE_AMPLITUDE_PARAMETERS = [
        "H0",
        "I21",
        "I31",
        "LAMBDA",
        "COSTHETA",
        "THETA",
        "Q22",
        "DIST",
    ]

    # the set of potential non-GR "amplitude" parameters
    NONGR_AMPLITUDE_PARAM = [
        "PHI01TENSOR",
        "PHI02TENSOR",
        "PHI01VECTOR",
        "PHI02VECTOR",
        "PHI01SCALAR",
        "PHI02SCALAR",
        "PSI1TENSOR",
        "PSI2TENSOR",
        "PSI1VECTOR",
        "PSI2VECTOR",
        "PSI1SCALAR",
        "PSI2SCALAR",
        "H1PLUS",
        "H2PLUS",
        "H1CROSS",
        "H2CROSS",
        "H1VECTORX",
        "H2VECTORX",
        "H1VECTORY",
        "H2VECTORY",
        "H1SCALARB",
        "H2SCALARB",
        "H1SCALARL",
        "H2SCALARL",
    ]

    # the set of positional parameters
    POSITIONAL_PARAMETERS = ["RAJ", "DECJ", "RA", "DEC", "PMRA", "PMDEC", "POSEPOCH"]

    # the set of potential binary parameters
    BINARY_PARAMS = [
        "PB",
        "ECC",
        "EPS1",
        "EPS2",
        "T0",
        "TASC",
        "A1",
        "OM",
        "PB_2",
        "ECC_2",
        "T0_2",
        "A1_2",
        "OM_2",
        "PB_3",
        "ECC_3",
        "T0_3",
        "A1_3",
        "OM_3",
        "XPBDOT",
        "EPS1DOT",
        "EPS2DOT",
        "OMDOT",
        "GAMMA",
        "PBDOT",
        "XDOT",
        "EDOT",
        "SINI",
        "DR",
        "DTHETA",
        "A0",
        "B0",
        "MTOT",
        "M2",
        "FB",
    ]

    # the parameters that are held as vectors
    VECTOR_PARAMS = ["F", "GLEP", "GLPH", "GLF0", "GLF1", "GLF2", "GLF0D", "GLTD", "FB"]

    def __init__(
        self, data, priors, par=None, det=None, likelihood="studentst", numba=False
    ):

        super().__init__(dict())  # initialise likelihood class

        # set the data
        if isinstance(data, HeterodynedData):
            # convert HeterodynedData to MultiHeterodynedData object
            if data.detector is None:
                raise ValueError("'data' must contain a named detector.")

            if data.par is None:
                raise ValueError("'data' must contain a heterodyne parameter file.")

            self.data = MultiHeterodynedData(data)
        elif isinstance(data, MultiHeterodynedData):
            if None in data.pars:
                raise ValueError(
                    "A data set does not have an associated "
                    "heterodyned parameter set."
                )

            self.data = data
        else:
            raise TypeError(
                "'data' must be a HeterodynedData or MultiHeterodynedData object."
            )

        if not isinstance(priors, bilby.core.prior.PriorDict):
            raise TypeError("Prior must be a bilby PriorDict")
        else:
            self.priors = priors

        # set the likelihood function
        self.likelihood = likelihood
        self._noise_log_likelihood = -np.inf  # initialise noise log likelihood
        self.numba = numba

        # check if prior includes (non-DeltaFunction) non-amplitude parameters,
        # i.e., will the phase evolution have to be included in the search,
        # or binary parameters
        self.include_phase = False
        self.include_binary = False
        self.include_glitch = False
        self.update_ssb = False
        for key in self.priors:
            if (
                key.upper()
                not in self.AMPLITUDE_PARAMS
                + self.BINARY_PARAMS
                + self.POSITIONAL_PARAMETERS
            ) and not self._is_vector_param(key.upper()):
                raise ValueError("Unknown parameter '{}' being used!".format(key))

            if key.upper() not in self.AMPLITUDE_PARAMS:
                if self.priors[key].is_fixed:
                    # check if it's the same value as the par files
                    for het in self.data:
                        if self._is_vector_param(key.upper()):
                            name, idx = self._vector_param_name_index(key.upper())
                            checkval = het.par[name][idx]
                        else:
                            checkval = het.par[key.upper()]

                        if checkval != self.priors[key].peak:
                            self.include_phase = True
                            if key.upper() in self.BINARY_PARAMS:
                                self.include_binary = True
                            elif key.upper() in self.POSITIONAL_PARAMETERS:
                                self.update_ssb = True
                            elif len(key) > 2:
                                if key.upper()[0:2] == "GL":
                                    self.include_glitch = True
                            break
                else:
                    self.include_phase = True
                    if key.upper() in self.BINARY_PARAMS:
                        self.include_binary = True
                    elif key.upper() in self.POSITIONAL_PARAMETERS:
                        self.update_ssb = True
                    elif len(key) > 2:
                        if key.upper()[0:2] == "GL":
                            self.include_glitch = True

        # check if any non-GR "amplitude" parameters are set
        self.nonGR = False
        for key in self.priors:
            if key.upper() in self.NONGR_AMPLITUDE_PARAM:
                self.nonGR = True

        # set up signal model classes
        self.models = []
        self.basepars = []
        for het in self.data:
            self.models.append(
                HeterodynedCWSimulator(
                    het.par,
                    het.detector,
                    het.times,
                    earth_ephem=het.ephemearth,
                    sun_ephem=het.ephemsun,
                )
            )
            # copy of heterodyned parameters
            newpar = PulsarParametersPy()
            for item in het.par.items():
                newpar[item[0]] = item[1]
            self.basepars.append(newpar)

        # if phase evolution is not in the model set the pre-summed products
        # of the data and antenna patterns
        if not self.include_phase:
            self.dot_products()

    @property
    def likelihood(self):
        """
        The the string containing the 'type' of likelihood function to use:
        'studentst' or 'gaussian'.
        """

        return self.__likelihood

    @likelihood.setter
    def likelihood(self, likelihood):
        # make sure the likelihood is of the allowed type
        if likelihood.lower() not in [
            "studentst",
            "students-t",
            "studentt",
            "gaussian",
            "normal",
        ]:
            raise ValueError("Likelihood must be 'studentst' or 'gaussian'.")

        if likelihood.lower() in ["studentst", "students-t", "studentt"]:
            self.__likelihood = "studentst"
        else:
            self.__likelihood = "gaussian"

    def dot_products(self):
        """
        Calculate the (noise-weighted) dot products of the data and the
        antenna pattern functions (see Appendix C of [1]_). E.g., for the data this is
        the real value

        .. math::

           (d/\\sigma) \\cdot (d^*/\\sigma) = \\sum_i \\frac{d_id_i^*}{\\sigma_i^2}
           \\equiv \\sum_i \\frac{\\Re{(d)}^2 + \\Im{(d)}^2}{\\sigma_i^2}.

        For the antenna patterns, for example :math:`F_+` and
        :math:`F_{\\times}`, we would have

        .. math::

           (F_+/\\sigma) \\cdot (F_+/\\sigma) = \\sum_i \\frac{{F_+}_i^2}{\\sigma_i^2},

           (F_\\times/\\sigma) \\cdot (F_\\times/\\sigma) = \\sum_i \\frac{{F_\\times}_i^2}{\\sigma_i^2},

           (F_+/\\sigma) \\cdot (F_{\\times}/\\sigma) = \\sum_i \\frac{{F_+}_i{F_\\times}_i}{\\sigma_i^2},

           (d/\\sigma) \\cdot (F_+/\\sigma) = \\sum_i \\frac{d_i{F_+}_i}{\\sigma_i^2},

           (d/\\sigma) \\cdot (F_\\times/\\sigma) = \\sum_i \\frac{d_i{F_\\times}_i}{\\sigma_i^2},

        For non-GR signals, also involving the vector and scalar modes, there
        are similar products.

        References
        ----------

        .. [1] M. Pitkin, M. Isi, J. Veitch & G. Woan, `arXiv:1705.08978v1
           <https:arxiv.org/abs/1705.08978v1>`_, 2017.
        """

        self.products = []  # list of products for each data set

        # loop over HeterodynedData and model functions
        for data, model in zip(self.data, self.models):
            if not self.numba:
                self.products.append(dict())
            else:
                dreal = numbadict.empty(
                    key_type=types.unicode_type, value_type=types.float64[:]
                )
                dcomplex = numbadict.empty(
                    key_type=types.unicode_type, value_type=types.complex128[:]
                )
                self.products.append([dreal, dcomplex])

            # initialise arrays in dictionary (Tp and Tc are the tensor plus
            # and cross, Vx and Vy are the vector "x" and "y", Sl and Sb are
            # the scalar longitudinal and breathing, d is the data)
            knames = ["d", "Tp", "Tc", "Vx", "Vy", "Sb", "Sl"]
            for i, a in enumerate(knames):
                for b in knames[::-1][: len(knames) - i]:
                    kname = a + "dot" + b
                    if "d" in [a, b] and a != b:
                        dtype = "complex128"
                    else:
                        dtype = "f8"

                        # set "whitened" versions of the antenna pattern product
                        # for SNR calculations, for when using the Students-t
                        # likelihood
                        if "d" not in [a, b]:
                            if not self.numba:
                                self.products[-1][kname + "White"] = np.zeros(
                                    data.num_chunks, dtype=dtype
                                )
                            else:
                                self.products[-1][0][kname + "White"] = np.zeros(
                                    data.num_chunks, dtype=dtype
                                )

                    if not self.numba:
                        self.products[-1][kname] = np.zeros(
                            data.num_chunks, dtype=dtype
                        )
                    else:
                        if dtype == "f8":
                            self.products[-1][0][kname] = np.zeros(
                                data.num_chunks, dtype=dtype
                            )
                        else:
                            self.products[-1][1][kname] = np.zeros(
                                data.num_chunks, dtype=dtype
                            )

            # loop over "chunks" into which each data set has been split
            for i, cpidx, cplen in zip(
                range(data.num_chunks), data.change_point_indices, data.chunk_lengths
            ):
                # set the noise standard deviation for a Gaussian likelihood
                stdstrue = data.stds[cpidx : cpidx + cplen]
                stds = stdstrue if self.likelihood == "gaussian" else 1.0

                # get the interpolated response functions
                t0 = float(model.resp.t0)

                # interpolation times
                ftimes = np.arange(
                    0.0, lal.DAYSID_SI, lal.DAYSID_SI / model.resp.ntimebins
                )
                inttimes = data.times[cpidx : cpidx + cplen].value - t0

                # dictionary of chunk data and antenna responses
                rs = dict()
                rs["d"] = data.data[cpidx : cpidx + cplen]
                rs["Tp"] = np.interp(
                    inttimes, ftimes, model.resp.fplus.data, period=lal.DAYSID_SI
                )
                rs["Tc"] = np.interp(
                    inttimes, ftimes, model.resp.fcross.data, period=lal.DAYSID_SI
                )
                rs["Vx"] = np.interp(
                    inttimes, ftimes, model.resp.fx.data, period=lal.DAYSID_SI
                )
                rs["Vy"] = np.interp(
                    inttimes, ftimes, model.resp.fy.data, period=lal.DAYSID_SI
                )
                rs["Sb"] = np.interp(
                    inttimes, ftimes, model.resp.fb.data, period=lal.DAYSID_SI
                )
                rs["Sl"] = np.interp(
                    inttimes, ftimes, model.resp.fl.data, period=lal.DAYSID_SI
                )

                # get all required combinations of responses and data
                for j, a in enumerate(knames):
                    for b in knames[::-1][: len(knames) - j]:
                        kname = a + "dot" + b
                        if not self.numba:
                            if kname == "ddotd":
                                # complex conjugate dot product for data
                                self.products[-1][kname][i] = np.vdot(
                                    rs[a] / stds, rs[b] / stds
                                ).real
                            else:
                                self.products[-1][kname][i] = np.dot(
                                    rs[a] / stds, rs[b] / stds
                                )

                            if "d" not in [a, b]:
                                # get "whitened" versions for Students-t likelihood
                                self.products[-1][kname + "White"][i] = np.dot(
                                    rs[a] / stdstrue, rs[b] / stdstrue
                                )
                        else:
                            if kname == "ddotd":
                                # complex conjugate dot product for data
                                self.products[-1][0][kname][i] = np.vdot(
                                    rs[a] / stds, rs[b] / stds
                                ).real
                            else:
                                if "d" in [a, b]:
                                    self.products[-1][1][kname][i] = np.dot(
                                        rs[a] / stds, rs[b] / stds
                                    )
                                else:
                                    self.products[-1][0][kname][i] = np.dot(
                                        rs[a] / stds, rs[b] / stds
                                    )

                            if "d" not in [a, b]:
                                # get "whitened" versions for Students-t likelihood
                                self.products[-1][0][kname + "White"][i] = np.dot(
                                    rs[a] / stdstrue, rs[b] / stdstrue
                                )

    def log_likelihood(self):
        """
        The log-likelihood function.
        """

        loglikelihood = 0.0  # the log likelihood value

        # loop over the data and models
        for data, model, prods, par in zip(
            self.data, self.models, self.products, self.basepars
        ):

            # update parameters in the base par
            for pname, pval in self.parameters.items():
                if self._is_vector_param(pname.upper()):
                    name = self._vector_param_name_index(pname.upper())[0]
                    par[name] = self._parse_vector_param(par, pname.upper(), pval)
                else:
                    # make sure values are floats
                    par[pname.upper()] = float(pval)

                    if pname.upper() in self.SOURCE_AMPLITUDE_PARAMETERS:
                        # reset waveform parameters (otherwise these can
                        # potentially be given by the previous value if the
                        # new value is zero). NOTE: in the future it might be
                        # better to pass a copy of the PulsarParameters
                        # structure to the model function, so that this is not
                        # problem
                        par["C21"] = 0.0
                        par["C22"] = 0.0
                        par["PHI22"] = 0.0
                        par["PHI21"] = 0.0

            # calculate the model
            m = model.model(
                par,
                usephase=self.include_phase,
                updateSSB=self.update_ssb,
                updateBSB=self.include_binary,
                updateglphase=self.include_glitch,
                freqfactor=data.freq_factor,
            )

            # calculate the likelihood
            if self.numba:
                loglikelihood += self._log_likelihood_numba(
                    data.data,
                    data.num_chunks,
                    np.asarray(data.change_point_indices),
                    np.asarray(data.chunk_lengths),
                    m,
                    self.likelihood,
                    prods[0],
                    prods[1],
                    data.stds,
                    self.nonGR,
                    self.include_phase,
                )
            else:
                # loop over stationary data chunks
                for i, cpidx, cplen in zip(
                    range(data.num_chunks),
                    data.change_point_indices,
                    data.chunk_lengths,
                ):
                    # likelihood without pre-summed products
                    if self.likelihood == "gaussian":
                        stds = data.stds[cpidx : cpidx + cplen]
                    else:
                        stds = 1.0

                    if self.include_phase:
                        # data and model for chunk
                        dd = data.data[cpidx : cpidx + cplen] / stds
                        mm = m[cpidx : cpidx + cplen] / stds

                        summodel = np.vdot(mm, mm).real
                        sumdatamodel = np.vdot(dd, mm).real
                    else:
                        # likelihood with pre-summed products
                        mp = m[0]  # tensor plus model component
                        mc = m[1]  # tensor cross model component

                        summodel = (
                            prods["TpdotTp"][i] * (mp.real ** 2 + mp.imag ** 2)
                            + prods["TcdotTc"][i] * (mc.real ** 2 + mc.imag ** 2)
                            + 2.0
                            * prods["TpdotTc"][i]
                            * (mp.real * mc.real + mp.imag * mc.imag)
                        )

                        sumdatamodel = (
                            prods["ddotTp"][i].real * mp.real
                            + prods["ddotTp"][i].imag * mp.imag
                            + prods["ddotTc"][i].real * mc.real
                            + prods["ddotTc"][i].imag * mc.imag
                        )

                        if self.nonGR:
                            # non-GR amplitudes
                            mx = m[2]
                            my = m[3]
                            mb = m[4]
                            ml = m[5]

                            summodel += (
                                prods["VxdotVx"][i] * (mx.real ** 2 + mx.imag ** 2)
                                + prods["VydotVy"][i] * (my.real ** 2 + my.imag ** 2)
                                + prods["SbdotSb"][i] * (mb.real ** 2 + mb.imag ** 2)
                                + prods["SldotSl"][i] * (ml.real ** 2 + ml.imag ** 2)
                                + 2.0
                                * (
                                    prods["TpdotVx"][i]
                                    * (mp.real * mx.real + mp.imag * mx.imag)
                                    + prods["TpdotVy"][i]
                                    * (mp.real * my.real + mp.imag * my.imag)
                                    + prods["TpdotSb"][i]
                                    * (mp.real * mb.real + mp.imag * mb.imag)
                                    + prods["TpdotSl"][i]
                                    * (mp.real * ml.real + mp.imag * ml.imag)
                                    + prods["TcdotVx"][i]
                                    * (mc.real * mx.real + mc.imag * mx.imag)
                                    + prods["TcdotVy"][i]
                                    * (mc.real * my.real + mc.imag * my.imag)
                                    + prods["TcdotSb"][i]
                                    * (mc.real * mb.real + mc.imag * mb.imag)
                                    + prods["TcdotSl"][i]
                                    * (mc.real * ml.real + mc.imag * ml.imag)
                                    + prods["VxdotVy"][i] * (mx.real * my.real)
                                    + (mx.imag * my.imag)
                                    + prods["VxdotSb"][i]
                                    * (mx.real * mb.real + mx.imag * mb.imag)
                                    + prods["VxdotSl"][i]
                                    * (mx.real * ml.real + mx.imag * ml.imag)
                                    + prods["VydotSb"][i]
                                    * (my.real * mb.real + my.imag * mb.imag)
                                    + prods["VydotSl"][i]
                                    * (my.real * ml.real + my.imag * ml.imag)
                                    + prods["SbdotSl"][i]
                                    * (mb.real * ml.real + mb.imag * ml.imag)
                                )
                            )

                            sumdatamodel += (
                                prods["ddotVx"][i].real * mx.real
                                + prods["ddotVx"][i].imag * mx.imag
                                + prods["ddotVy"][i].real * my.real
                                + prods["ddotVy"][i].imag * my.imag
                                + prods["ddotSb"][i].real * mb.real
                                + prods["ddotSb"][i].imag * mb.imag
                                + prods["ddotSl"][i].real * ml.real
                                + prods["ddotSl"][i].imag * ml.imag
                            )

                    # compute "Chi-squared"
                    chisquare = prods["ddotd"][i] - 2.0 * sumdatamodel + summodel

                    if self.likelihood == "gaussian":
                        loglikelihood += 0.5 * chisquare
                        # normalisation
                        loglikelihood -= np.log(lal.TWOPI * stds[0] ** 2)
                    else:
                        loglikelihood += (
                            logfactorial(cplen - 1)
                            - lal.LN2
                            - cplen * lal.LNPI
                            - cplen * np.log(chisquare)
                        )

        return loglikelihood

    @staticmethod
    @jit(nopython=True)
    def _log_likelihood_numba(
        data,
        datanchunks,
        datacps,
        datacls,
        model,
        likelihood="studentst",
        productsreal=None,
        productscomp=None,
        datastds=None,
        nonGR=False,
        includephase=False,
    ):
        """
        This is a version of the standard inner loop of
        :meth:`cwinpy.TargetedPulsarLikelihood.log_likelihood` that used the
        `numba <https://numba.pydata.org/>`_ JIT package to provide some
        speed-up.

        Parameters
        ----------
        data: array_like
            A complex :class:`numpy.ndarray` containing data from a single
            detector.
        datanchunks: int
            The number of chunks into which the data has been split.
        datacps: array_like
            A :class:`numpy.ndarray` containing the change point indices for
            each chunk.
        datacls: array_like
            A :class:`numpy.ndarray` containing lengths of each of the chunks.
        model: array_like
            A complex :class:`numpy.ndarray` containing the signal model.
        likelihood: str
            A string stating whether to use the Student's-t or Gaussian
            likelihood.
        productsreal: dict
            A numba :class:`~numba.typed.Dict` containing
            :class:`numpy.ndarray`'s for model component dot products.
        productscomp: dict
            A numba :class:`~numba.typed.Dict` containing
            :class:`numpy.ndarray`'s for data and model component dot products.
        datastds: array_like
            A :class:`numpy.ndarray` containing data standard deviations.
        nonGR: bool
            A flag to set whether using a non-GR signal.
        includephase: bool
            A flag specifying whether the model has a varying phase evolution.

        Returns
        -------
        loglikelihood: float
            The log-likelihood function calculated for that data set.
        """

        loglikelihood = 0.0

        # calculate the likelihood
        for i, cpidx, cplen in zip(range(datanchunks), datacps, datacls):
            # loop over stationary data chunks

            # likelihood without pre-summed products
            if likelihood == "gaussian":
                stds = datastds[cpidx : cpidx + cplen]
            else:
                stds = np.ones(cplen)

            if includephase:
                # data and model for chunk
                dd = data[cpidx : cpidx + cplen] / stds
                mm = model[cpidx : cpidx + cplen] / stds

                summodel = np.vdot(mm, mm).real
                sumdatamodel = np.vdot(dd, mm).real
            else:
                # likelihood with pre-summed products
                mp = model[0]  # tensor plus model component
                mc = model[1]  # tensor cross model component

                summodel = (
                    productsreal["TpdotTp"][i] * (mp.real ** 2 + mp.imag ** 2)
                    + productsreal["TcdotTc"][i] * (mc.real ** 2 + mc.imag ** 2)
                    + 2.0
                    * productsreal["TpdotTc"][i]
                    * (mp.real * mc.real + mp.imag * mc.imag)
                )

                sumdatamodel = (
                    productscomp["ddotTp"][i].real * mp.real
                    + productscomp["ddotTp"][i].imag * mp.imag
                    + productscomp["ddotTc"][i].real * mc.real
                    + productscomp["ddotTc"][i].imag * mc.imag
                )

                if nonGR:
                    # non-GR amplitudes
                    mx = model[2]
                    my = model[3]
                    mb = model[4]
                    ml = model[5]

                    summodel += (
                        productsreal["VxdotVx"][i] * (mx.real ** 2 + mx.imag ** 2)
                        + productsreal["VydotVy"][i] * (my.real ** 2 + my.imag ** 2)
                        + productsreal["SbdotSb"][i] * (mb.real ** 2 + mb.imag ** 2)
                        + productsreal["SldotSl"][i] * (ml.real ** 2 + ml.imag ** 2)
                        + 2.0
                        * (
                            productsreal["TpdotVx"][i]
                            * (mp.real * mx.real + mp.imag * mx.imag)
                            + productsreal["TpdotVy"][i]
                            * (mp.real * my.real + mp.imag * my.imag)
                            + productsreal["TpdotSb"][i]
                            * (mp.real * mb.real + mp.imag * mb.imag)
                            + productsreal["TpdotSl"][i]
                            * (mp.real * ml.real + mp.imag * ml.imag)
                            + productsreal["TcdotVx"][i]
                            * (mc.real * mx.real + mc.imag * mx.imag)
                            + productsreal["TcdotVy"][i]
                            * (mc.real * my.real + mc.imag * my.imag)
                            + productsreal["TcdotSb"][i]
                            * (mc.real * mb.real + mc.imag * mb.imag)
                            + productsreal["TcdotSl"][i]
                            * (mc.real * ml.real + mc.imag * ml.imag)
                            + productsreal["VxdotVy"][i] * (mx.real * my.real)
                            + (mx.imag * my.imag)
                            + productsreal["VxdotSb"][i]
                            * (mx.real * mb.real + mx.imag * mb.imag)
                            + productsreal["VxdotSl"][i]
                            * (mx.real * ml.real + mx.imag * ml.imag)
                            + productsreal["VydotSb"][i]
                            * (my.real * mb.real + my.imag * mb.imag)
                            + productsreal["VydotSl"][i]
                            * (my.real * ml.real + my.imag * ml.imag)
                            + productsreal["SbdotSl"][i]
                            * (mb.real * ml.real + mb.imag * ml.imag)
                        )
                    )

                    sumdatamodel += (
                        productscomp["ddotVx"][i].real * mx.real
                        + productscomp["ddotVx"][i].imag * mx.imag
                        + productscomp["ddotVy"][i].real * my.real
                        + productscomp["ddotVy"][i].imag * my.imag
                        + productscomp["ddotSb"][i].real * mb.real
                        + productscomp["ddotSb"][i].imag * mb.imag
                        + productscomp["ddotSl"][i].real * ml.real
                        + productscomp["ddotSl"][i].imag * ml.imag
                    )

            # compute "Chi-squared"
            chisquare = productsreal["ddotd"][i] - 2.0 * sumdatamodel + summodel

            if likelihood == "gaussian":
                loglikelihood += 0.5 * chisquare
                # normalisation
                loglikelihood -= np.log(lal.TWOPI * stds[0] ** 2)
            else:
                loglikelihood += (
                    logfactorial(cplen - 1)
                    - lal.LN2
                    - cplen * lal.LNPI
                    - cplen * np.log(chisquare)
                )

        return loglikelihood

    def noise_log_likelihood(self):
        """
        The log-likelihood for the data being consistent with the noise model,
        i.e., when the signal is zero. See Equations 14 and 15 of [1]_.

        Returns
        -------
        float:
            The noise-only log-likelihood

        References
        ----------

        .. [1] M. Pitkin, M. Isi, J. Veitch & G. Woan, `arXiv:1705.08978v1
           <https:arxiv.org/abs/1705.08978v1>`_, 2017.
        """

        if np.isfinite(self._noise_log_likelihood):
            # return cached version of the noise log likelihood
            return self._noise_log_likelihood

        self._noise_log_likelihood = 0.0  # the log likelihood value

        # loop over the data and models
        for data, prods in zip(self.data, self.products):
            # calculate the likelihood
            for i, cpidx, cplen in zip(
                range(data.num_chunks), data.change_point_indices, data.chunk_lengths
            ):
                # loop over stationary data chunks
                if self.numba:
                    ddotd = prods[0]["ddotd"][i]
                else:
                    ddotd = prods["ddotd"][i]
                if self.likelihood == "gaussian":
                    self._noise_log_likelihood += 0.5 * ddotd
                    # normalisation
                    self._noise_log_likelihood -= np.log(lal.TWOPI * data.vars[cpidx])
                else:
                    self._noise_log_likelihood += (
                        logfactorial(cplen - 1)
                        - lal.LN2
                        - cplen * lal.LNPI
                        - cplen * np.log(ddotd)
                    )

        return self._noise_log_likelihood

    def _is_vector_param(self, name):
        """
        Check if a parameter is a vector parameter.
        """

        # check for integers in name
        intvals = re.findall(r"\d+", name)
        if len(intvals) == 0:
            return False

        # strip out any underscores from name and remove trailing index
        noscores = re.sub("_", "", name)[: -len(intvals[-1])]

        if noscores in self.VECTOR_PARAMS:
            return True
        else:
            return False

    def _vector_param_name_index(self, name):
        """
        Get the vector parameter name (stripped of the position)
        """

        intvals = re.findall(r"\d+", name)

        # strip out any underscores from name and remove trailing index
        noscores = re.sub("_", "", name)[: -len(intvals[-1])]

        # glitch values start from 1 so subtract 1 from pos
        if name[:2] == "GL":
            intvals[-1] -= 1

        return (noscores, intvals[-1])

    def _parse_vector_param(self, par, name, value):
        """
        Set a vector parameter with the given single value at the place
        specified in the `name`.
        """

        vname, vpos = self._vector_param_name_index(name)
        vec = par[vname]
        vec[vpos] = value

        return vec



"""
Test script for utils.py function.
"""

import os

import numpy as np
import pytest
from astropy import units as u
from cwinpy.utils import (
    ellipticity_to_q22,
    gcd_array,
    get_psr_name,
    initialise_ephemeris,
    int_to_alpha,
    is_par_file,
    logfactorial,
    q22_to_ellipticity,
)
from lalpulsar.PulsarParametersWrapper import PulsarParametersPy


def test_logfactorial():
    """
    Test log factorial function
    """

    a = 3
    assert logfactorial(a) == np.log(3 * 2 * 1)

    a = 3.0
    assert logfactorial(a) == np.log(3 * 2 * 1)


def test_gcd_array():
    """
    Test greatest common divisor function.
    """

    a = 1  # non-list value
    with pytest.raises(TypeError):
        gcd_array(a)

    a = [1]  # single value
    with pytest.raises(ValueError):
        gcd_array(a)

    a = [5, 25, 90]
    assert gcd_array(a) == 5


def test_int_to_alpha():
    """
    Test integer to alphabetical string conversion.
    """

    pos = 2.3
    with pytest.raises(TypeError):
        int_to_alpha(pos)

    pos = -1
    with pytest.raises(ValueError):
        int_to_alpha(pos)

    assert int_to_alpha(1) == "A"
    assert int_to_alpha(1, case="lower") == "a"
    assert int_to_alpha(26) == "Z"
    assert int_to_alpha(26, case="lower") == "z"
    assert int_to_alpha(27) == "AA"
    assert int_to_alpha(28) == "AB"
    assert int_to_alpha(200) == "GR"
    assert int_to_alpha(1000) == "ALL"


def test_is_par_file():
    """
    Test failure of is_par_file.
    """

    assert is_par_file("blah_blah_blah") is False

    # test par files that don't contain required attributes
    brokenpar = "broken.par"

    values = {
        "F": [100.0],
        "RAJ": 0.1,
        "DECJ": -0.1,
        "PSRJ": "J0101-0101",
    }

    for leavekey in list(values.keys()):
        keys = list(values.keys())
        psr = PulsarParametersPy()
        for key in keys:
            if key != leavekey:
                psr[key] = values[key]

        psr.pp_to_par(brokenpar)

        assert is_par_file(brokenpar) is False

        os.remove(brokenpar)


def test_get_psr_name():
    """
    Test extraction of pulsar name.
    """

    for item, name in zip(
        ["PSRJ", "PSRB", "PSR", "NAME"],
        ["J0123+1234", "B0124+12", "J0123+1234", "B0124+12"],
    ):
        psr = PulsarParametersPy()
        psr[item] = name

        assert get_psr_name(psr) == name


def test_ellipticity_to_q22():
    """
    Test ellipticity conversion to mass quadrupole.
    """

    epsilon = [1e-9, 1e-8]
    expected_q22 = np.array([1e29, 1e30]) * np.sqrt(15.0 / (8.0 * np.pi))
    q22 = ellipticity_to_q22(epsilon[0])

    assert np.isclose(q22, expected_q22[0])

    # test units
    q22units = ellipticity_to_q22(epsilon[0], units=True)

    assert np.isclose(q22units.value, expected_q22[0])
    assert q22units.unit == u.Unit("kg m2")

    # test array like
    q22 = ellipticity_to_q22(epsilon)

    assert len(q22) == len(epsilon)
    assert np.allclose(q22, expected_q22)


def test_q22_to_ellipticity_to_q22():
    """
    Test mass quadrupole conversion to ellipticity.
    """

    q22 = [1e29, 1e30]
    expected_epsilon = np.array([1e-9, 1e-8]) / np.sqrt(15.0 / (8.0 * np.pi))
    epsilon = q22_to_ellipticity(q22[0])

    assert np.isclose(epsilon, expected_epsilon[0])

    # test array like
    epsilon = q22_to_ellipticity(q22)

    assert len(q22) == len(epsilon)
    assert np.allclose(epsilon, expected_epsilon)

    # test no unit
    epsilon = q22_to_ellipticity(q22[0] * u.kg * u.m ** 2)

    assert np.isclose(epsilon, expected_epsilon[0])
    assert not hasattr(epsilon, "unit")


def test_initialise_ephemeris():
    """
    Test reading of ephemeris files.
    """

    with pytest.raises(ValueError):
        initialise_ephemeris(units="lhfld")

    with pytest.raises(IOError):
        initialise_ephemeris(
            earthfile="jksgdksg", sunfile="lhlbca", timefile="lshdldgls"
        )

    with pytest.raises(IOError):
        initialise_ephemeris(
            earthfile="jksgdksg", sunfile="lhlbca", timefile="lshdldgls"
        )

    with pytest.raises(IOError):
        initialise_ephemeris(timefile="lshdldgls")

    edat, tdat = initialise_ephemeris()

    assert edat.nentriesE == 175322
    assert edat.nentriesS == 17534
    assert edat.dtEtable == 7200.0
    assert edat.dtStable == 72000.0
    assert edat.etype == 2

    assert tdat.nentriesT == 87660
    assert tdat.dtTtable == 14400.0



"""
Output various comparitors between lalapps_pulsar_parameter_estimation_nested
runs and cwinpy runs
"""

import os

import bilby
import cwinpy
import h5py
import numpy as np
from bilby.core.result import read_in_result
from lalinference import LALInferenceHDF5PosteriorSamplesDatasetName
from lalinference.io import read_samples
from scipy.spatial.distance import jensenshannon
from scipy.stats import combine_pvalues, gaussian_kde, ks_2samp

# comparison rst table information
FILETEXT = """\
.. csv-table:: Evidence table
   :widths: auto
   :header: "Method", ":math:`\\\\ln{{(Z)}}`", ":math:`\\\\ln{{(Z)}}` noise", ":math:`\\\\ln{{}}` Odds"

   "``lalapps_pulsar_parameter_estimation_nested``", "{0:.3f}", "{1:.3f}", "{2:.3f}±{3:.3f}"
   "``cwinpy_pe``", "{4:.3f}", "{5:.3f}", "{6:.3f}±{7:.3f}"
   "``cwinpy_pe`` (grid)", "{8}", "", "{9}"

.. csv-table:: Parameter table
   :widths: auto
   :header: "Method", ":math:`h_0`", ":math:`\\\\phi_0` (rad)", ":math:`\\\\psi` (rad)", ":math:`\\\\cos{{\\\\iota}}`"

   "``lalapps_pulsar_parameter_estimation_nested``", "{10:.2f}±{11:.2f}×10\ :sup:`{12:d}`", "{13:.2f}±{14:.2f}", "{15:.2f}±{16:.2f}", "{17:.2f}±{18:.2f}"
   "{19:d}% credible intervals", "[{20:.2f}, {21:.2f}]×10\ :sup:`{22:d}`", "[{23:.2f}, {24:.2f}]", "[{25:.2f}, {26:.2f}]", "[{27:.2f}, {28:.2f}]"
   "``cwinpy_pe``", "{29:.2f}±{30:.2f}×10\ :sup:`{31:d}`", "{32:.2f}±{33:.2f}", "{34:.2f}±{35:.2f}", "{36:.2f}±{37:.2f}"
   "{38:d}% credible intervals", "[{39:.2f}, {40:.2f}]×10\ :sup:`{41:d}`", "[{42:.2f}, {43:.2f}]", "[{44:.2f}, {45:.2f}]", "[{46:.2f}, {47:.2f}]"

.. csv-table:: Maximum a-posteriori
   :widths: auto
   :header: "Method", ":math:`h_0`", ":math:`\\\\phi_0` (rad)", ":math:`\\\\psi` (rad)", ":math:`\\\\cos{{\\\\iota}}`", ":math:`\\\\ln{{(L)}}` max"

   "``lalapps_pulsar_parameter_estimation_nested``", "{48:.2f}×10\ :sup:`{49:d}`", "{50:.2f}", "{51:.2f}", "{52:.2f}", "{53:.2f}"
   "``cwinpy_pe``", "{54:.2f}×10\ :sup:`{55:d}`", "{56:.2f}", "{57:.2f}", "{58:.2f}", "{59:.2f}"

| Combined K-S test p-value: {60:.4f}
| Maximum Jensen-Shannon divergence: {61:.4f}

| CWInPy version: {62:s}
| bilby version: {63:s}
"""


# comparison rst table information for two harmonics
FILETEXTTWOHARMONICS = """\
.. csv-table:: Evidence table
   :widths: auto
   :header: "Method", ":math:`\\\\ln{{(Z)}}`", ":math:`\\\\ln{{(Z)}}` noise", ":math:`\\\\ln{{}}` Odds"

   "``lalapps_pulsar_parameter_estimation_nested``", "{0:.3f}", "{1:.3f}", "{2:.3f}±{3:.3f}"
   "``cwinpy_pe``", "{4:.3f}", "{5:.3f}", "{6:.3f}±{7:.3f}"

.. csv-table:: Parameter table
   :widths: auto
   :header: "Method", ":math:`C_{{21}}`", ":math:`C_{{22}}`", ":math:`\\\\Phi_{{21}}` (rad)", ":math:`\\\\Phi_{{22}}` (rad)", ":math:`\\\\psi` (rad)", ":math:`\\\\cos{{\\\\iota}}`"

   "``lalapps_pulsar_parameter_estimation_nested``", "{8:.2f}±{9:.2f}×10\ :sup:`{10:d}`", "{11:.2f}±{12:.2f}×10\ :sup:`{13:d}`", "{14:.2f}±{15:.2f}", "{16:.2f}±{17:.2f}", "{18:.2f}±{19:.2f}", "{20:.2f}±{21:.2f}"
   "{22:d}% credible intervals", "[{23:.2f}, {24:.2f}]×10\ :sup:`{25:d}`", "[{26:.2f}, {27:.2f}]×10\ :sup:`{28:d}`", "[{29:.2f}, {30:.2f}]", "[{31:.2f}, {32:.2f}]", "[{33:.2f}, {34:.2f}]", "[{35:.2f}, {36:.2f}]"
   "``cwinpy_pe``", "{37:.2f}±{38:.2f}×10\ :sup:`{39:d}`", "{40:.2f}±{41:.2f}×10\ :sup:`{42:d}`", "{43:.2f}±{44:.2f}", "{45:.2f}±{46:.2f}", "{47:.2f}±{48:.2f}", "{49:.2f}±{50:.2f}"
   "{51:d}% credible intervals", "[{52:.2f}, {53:.2f}]×10\ :sup:`{54:d}`", "[{55:.2f}, {56:.2f}]×10\ :sup:`{57:d}`", "[{58:.2f}, {59:.2f}]", "[{60:.2f}, {61:.2f}]", "[{62:.2f}, {63:.2f}]", "[{64:.2f}, {65:.2f}]"

.. csv-table:: Maximum a-posteriori
   :widths: auto
   :header: "Method", ":math:`C_{{21}}`", ":math:`C_{{22}}`", ":math:`\\\\Phi_{{21}}` (rad)", ":math:`\\\\Phi_{{22}}` (rad)", ":math:`\\\\psi` (rad)", ":math:`\\\\cos{{\\\\iota}}`", ":math:`\\\\ln{{(L)}}` max"

   "``lalapps_pulsar_parameter_estimation_nested``", "{66:.2f}×10\ :sup:`{67:d}`", "{68:.2f}×10\ :sup:`{69:d}`", "{70:.2f}", "{71:.2f}", "{72:.2f}", "{73:.2f}", "{74:.2f}"
   "``cwinpy_pe``", "{75:.2f}×10\ :sup:`{76:d}`", "{77:.2f}×10\ :sup:`{78:d}`", "{79:.2f}", "{80:.2f}", "{81:.2f}", "{82:.2f}", "{83:.2f}"

| Combined K-S test p-value: {84:.4f}
| Maximum Jensen-Shannon divergence: {85:.4f}

| CWInPy version: {86:s}
| bilby version: {87:s}
"""


def credible_interval(samples, ci=0.90):
    # get the given percentage credible interval about the median
    return np.quantile(samples, [0.5 - 0.5 * ci, 0.5 + 0.5 * ci])


def comparisons(label, outdir, grid, priors, cred=0.9):
    """
    Perform comparisons of the evidence, parameter values, confidence
    intervals, and Kolmogorov-Smirnov test between samples produced with
    lalapps_pulsar_parameter_estimation_nested and cwinpy.
    """

    lppenfile = os.path.join(outdir, "{}_post.hdf".format(label))

    # get posterior samples
    post = read_samples(
        lppenfile, tablename=LALInferenceHDF5PosteriorSamplesDatasetName
    )

    # get uncertainty on ln(evidence)
    info = h5py.File(lppenfile)["lalinference"]["lalinference_nest"].attrs[
        "information_nats"
    ]
    nlive = h5py.File(lppenfile)["lalinference"]["lalinference_nest"].attrs[
        "number_live_points"
    ]
    evsig = h5py.File(lppenfile)["lalinference"]["lalinference_nest"].attrs[
        "log_evidence"
    ]
    evnoise = h5py.File(lppenfile)["lalinference"]["lalinference_nest"].attrs[
        "log_noise_evidence"
    ]
    everr = np.sqrt(info / nlive)  # the uncertainty on the evidence

    # read in cwinpy results
    result = read_in_result(outdir=outdir, label=label)

    # comparison file
    comparefile = os.path.join(outdir, "{}_compare.txt".format(label))

    # get grid-based evidence
    if grid is not None:
        grid_evidence = grid.log_evidence

    # set values to output
    values = 64 * [None]
    values[0:4] = evsig, evnoise, (evsig - evnoise), everr
    values[4:8] = (
        result.log_evidence,
        result.log_noise_evidence,
        result.log_bayes_factor,
        result.log_evidence_err,
    )
    if grid is not None:
        values[8] = "{0:.3f}".format(grid_evidence)
        values[9] = "{0:.3f}".format(grid_evidence - result.log_noise_evidence)
    else:
        values[8:10] = ("N/A", "N/A")  # no values supplied

    # output parameter means standard deviations, and credible intervals
    idx = 10
    for method in ["lalapps", "cwinpy"]:
        values[idx + 9] = int(cred * 100)
        for p in priors.keys():
            samples = post[p.upper()] if method == "lalapps" else result.posterior[p]

            # convert iota to cos(iota)
            if p == "iota":
                samples = np.cos(samples)

            mean = samples.mean()
            std = samples.std()
            low, high = credible_interval(samples, ci=cred)
            if p == "h0":
                exponent = int(np.floor(np.log10(mean)))
                values[idx] = mean / 10 ** exponent
                values[idx + 1] = std / 10 ** exponent
                values[idx + 2] = exponent
                values[idx + 10] = low / 10 ** exponent
                values[idx + 11] = high / 10 ** exponent
                values[idx + 12] = exponent
                idx += 3
            else:
                values[idx] = mean
                values[idx + 1] = std
                values[idx + 10] = low
                values[idx + 11] = high
                idx += 2
        idx += 10

    # output parameter maximum a-posteriori points
    maxidx = (
        result.posterior["log_likelihood"] + result.posterior["log_prior"]
    ).idxmax()
    maxidxlppen = (post["logL"] + post["logPrior"]).argmax()
    for method in ["lalapps", "cwinpy"]:
        for p in priors.keys():
            maxpval = (
                post[p.upper()][maxidxlppen]
                if method == "lalapps"
                else result.posterior[p][maxidx]
            )
            if p == "h0":
                exponent = int(np.floor(np.log10(maxpval)))
                values[idx] = maxpval / 10 ** exponent
                values[idx + 1] = exponent
                idx += 2
            else:
                values[idx] = maxpval
                idx += 1
        if result.use_ratio:
            # convert likelihood ratio back to likelihood
            values[idx] = (
                post["logL"][maxidxlppen]
                if method == "lalapps"
                else (
                    result.posterior["log_likelihood"][maxidx]
                    + result.log_noise_evidence
                )
            )
        else:
            values[idx] = (
                post["logL"][maxidxlppen]
                if method == "lalapps"
                else result.posterior["log_likelihood"][maxidx]
            )
        idx += 1

    # calculate the Kolmogorov-Smirnov test for each 1d marginalised distribution,
    # and the Jensen-Shannon divergence, from the two codes. Output the
    # combined p-value of the KS test statistic over all parameters, and the
    # maximum Jensen-Shannon divergence over all parameters.
    values[idx] = np.inf
    pvalues = []
    jsvalues = []
    for p in priors.keys():
        _, pvalue = ks_2samp(post[p.upper()], result.posterior[p])
        pvalues.append(pvalue)

        # calculate J-S divergence (use Gaussian KDE)
        bins = np.linspace(
            np.min([np.min(post[p.upper()]), np.min(result.posterior[p])]),
            np.max([np.max(post[p.upper()]), np.max(result.posterior[p])]),
            100,
        )

        hp = gaussian_kde(post[p.upper()]).pdf(bins)
        hq = gaussian_kde(result.posterior[p]).pdf(bins)
        jsvalues.append(jensenshannon(hp, hq) ** 2)

    values[idx] = combine_pvalues(pvalues)[1]
    idx += 1
    values[idx] = np.max(jsvalues)

    values[idx + 1] = cwinpy.__version__
    values[idx + 2] = bilby.__version__

    with open(comparefile, "w") as fp:
        fp.write(FILETEXT.format(*values))


def comparisons_two_harmonics(label, outdir, priors, cred=0.9):
    """
    Perform comparisons of the evidence, parameter values, confidence
    intervals, and Kolmogorov-Smirnov test between samples produced with
    lalapps_pulsar_parameter_estimation_nested and cwinpy.
    """

    lppenfile = os.path.join(outdir, "{}_post.hdf".format(label))

    # get posterior samples
    post = read_samples(
        lppenfile, tablename=LALInferenceHDF5PosteriorSamplesDatasetName
    )

    # get uncertainty on ln(evidence)
    info = h5py.File(lppenfile)["lalinference"]["lalinference_nest"].attrs[
        "information_nats"
    ]
    nlive = h5py.File(lppenfile)["lalinference"]["lalinference_nest"].attrs[
        "number_live_points"
    ]
    evsig = h5py.File(lppenfile)["lalinference"]["lalinference_nest"].attrs[
        "log_evidence"
    ]
    evnoise = h5py.File(lppenfile)["lalinference"]["lalinference_nest"].attrs[
        "log_noise_evidence"
    ]
    everr = np.sqrt(info / nlive)  # the uncertainty on the evidence

    # read in cwinpy results
    result = read_in_result(outdir=outdir, label=label)

    # comparison file
    comparefile = os.path.join(outdir, "{}_compare.txt".format(label))

    # set values to output
    values = 88 * [None]
    values[0:4] = evsig, evnoise, (evsig - evnoise), everr
    values[4:8] = (
        result.log_evidence,
        result.log_noise_evidence,
        result.log_bayes_factor,
        result.log_evidence_err,
    )

    # output parameter means standard deviations, and credible intervals
    idx = 8
    for method in ["lalapps", "cwinpy"]:
        values[idx + 14] = int(cred * 100)
        for p in priors.keys():
            samples = post[p.upper()] if method == "lalapps" else result.posterior[p]

            # convert iota to cos(iota)
            if p == "iota":
                samples = np.cos(samples)

            mean = samples.mean()
            std = samples.std()
            low, high = credible_interval(samples, ci=cred)
            if p in ["c21", "c22"]:
                exponent = int(np.floor(np.log10(mean)))
                values[idx] = mean / 10 ** exponent
                values[idx + 1] = std / 10 ** exponent
                values[idx + 2] = exponent
                values[idx + 15] = low / 10 ** exponent
                values[idx + 16] = high / 10 ** exponent
                values[idx + 17] = exponent
                idx += 3
            else:
                values[idx] = mean
                values[idx + 1] = std
                values[idx + 15] = low
                values[idx + 16] = high
                idx += 2
        idx += 15

    # output parameter maximum a-posteriori points
    maxidx = (
        result.posterior["log_likelihood"] + result.posterior["log_prior"]
    ).idxmax()
    maxidxlppen = (post["logL"] + post["logPrior"]).argmax()
    for method in ["lalapps", "cwinpy"]:
        for p in priors.keys():
            maxpval = (
                post[p.upper()][maxidxlppen]
                if method == "lalapps"
                else result.posterior[p][maxidx]
            )
            if p in ["c21", "c22"]:
                exponent = int(np.floor(np.log10(maxpval)))
                values[idx] = maxpval / 10 ** exponent
                values[idx + 1] = exponent
                idx += 2
            else:
                values[idx] = maxpval
                idx += 1
        if result.use_ratio:
            values[idx] = (
                post["logL"][maxidxlppen]
                if method == "lalapps"
                else (
                    result.posterior["log_likelihood"][maxidx]
                    + result.log_noise_evidence
                )
            )
        else:
            values[idx] = (
                post["logL"][maxidxlppen]
                if method == "lalapps"
                else result.posterior["log_likelihood"][maxidx]
            )
        idx += 1

    # calculate the Kolmogorov-Smirnov test for each 1d marginalised distribution,
    # and the Jensen-Shannon divergence, from the two codes. Output the
    # combined p-value of the KS test statistic over all parameters, and the
    # maximum Jensen-Shannon divergence over all parameters.
    values[idx] = np.inf
    pvalues = []
    jsvalues = []
    for p in priors.keys():
        _, pvalue = ks_2samp(post[p.upper()], result.posterior[p])
        pvalues.append(pvalue)

        # calculate J-S divergence
        bins = np.linspace(
            np.min([np.min(post[p.upper()]), np.min(result.posterior[p])]),
            np.max([np.max(post[p.upper()]), np.max(result.posterior[p])]),
            100,
        )

        hp, _ = np.histogram(post[p.upper()], bins=bins, density=True)
        hq, _ = np.histogram(result.posterior[p], bins=bins, density=True)
        jsvalues.append(jensenshannon(hp, hq) ** 2)

    values[idx] = combine_pvalues(pvalues)[1]
    idx += 1
    values[idx] = np.max(jsvalues)

    values[idx + 1] = cwinpy.__version__
    values[idx + 2] = bilby.__version__

    with open(comparefile, "w") as fp:
        fp.write(FILETEXTTWOHARMONICS.format(*values))
